/*
  YAKS, Optann, a Khepera simulator including a separate GA and ANN 
  (Genetic Algoritm, Artificial Neural Net).
  Copyright (C) 2000  Johan Carlsson (johanc@ida.his.se)
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 */

#include "ann.h"

/**
 * Get last activation value from all nodes with the priority of pri
 * @param pri The priority of the nodes to look for
 * @return A structure with (struct)->nodes and the (struct)->values, or NULL if no nodes where found.
 */

valuesS *Ann::getValuesFromNodesWithPriority(int pri){
  int nrOfSuchNodes;
  int indexB,indexE,indexL;
  valuesS* list;
  nrOfSuchNodes=indexB=indexE=indexL=-1;

  for(int i=0; i < nrOfNeurons && indexB==-1; i++){
    if(actOrder[i]->getPriority()==pri && indexB==-1){
      indexB=i;
      for(i=indexB; i < nrOfNeurons && indexE==-1; i++){
	if(actOrder[i]->getPriority()!=pri){
	  indexE=i;
	}
      }
    }
  }
  indexL=indexE-indexB;
  if(indexL>0){
    list = (valuesS*) malloc(sizeof(valuesS));
    list->nodes=indexL;
    list->values = (double*)malloc(indexL*sizeof(double));
    for(int i=0; i < indexL; i++){
      list->values[i]=actOrder[indexB+i]->getActivationValue();
    }
    return list;
  }
  else{
    return NULL;
  } 
}
/**
 * Save all network weights to file.
 * @see loadWeights()
 * @param fp A file descriptor pointing to a file open for writing.
 */
void Ann::saveWeights(FILE *fp) const{
  for(int i=0; i < nrOfNeurons; i++){
    fprintf(fp,"Neuron %d Links %d ",i,actOrder[i]->getNrOfLinks());
#ifdef BIT_WEIGHTS
    for(int w=0; w < actOrder[i]->getNrOfLinks(); w++){
      fprintf(fp,"%ld ",(long int)actOrder[i]->getLinkWeight(w));
    }
#else
    for(int w=0; w < actOrder[i]->getNrOfLinks(); w++){
      fprintf(fp,"%f ",actOrder[i]->getLinkWeight(w));
    }
#endif
    fprintf(fp,"\n");
  }
}

/**
 * Save all network activations to file.
 * @param fp A file descriptor pointing to a file open for writing.
 */
void Ann::saveActivations(FILE *fp) const{
  for(int i=0; i < nrOfNeurons; i++){
    fprintf(fp,"%f\t",actOrder[i]->getActivationValue());
  }
  fprintf(fp,"\n");
}
/**
 * Load network weights from file.
 * @see saveWeights()
 * @param fp A file descriptor pointing to a file open for reading.
 */
void Ann::loadWeights(FILE *fp){
  int bogusI=0,bogusL=0;
  int scanned=0;
  Weight weight=0;
  for(int i=0; i < nrOfNeurons; i++){
    scanned=fscanf(fp,"Neuron %d Links %d ",&bogusI,&bogusL);
    if(bogusL != actOrder[i]->getNrOfLinks() || scanned!=2){
      fprintf(stderr,"Ann::loadWeights Wohaaaa something is wrong !!!\n");
      exit(-1);
    }
    for(int w=0; w < actOrder[i]->getNrOfLinks(); w++){
#ifdef BIT_WEIGHTS
      long int weights=0;
      fscanf(fp,"%ld ",&weights);
      weight = (Weight) weights;
      actOrder[i]->setLinkWeight(w,weight);
      weights=0;
#else
      fscanf(fp,"%lf ",&weight);
      actOrder[i]->setLinkWeight(w,weight);
#endif
    }
    fscanf(fp,"\n");
  }
}

/**
 *Print the network architecture with weights and activation values to stdout.
 */
void Ann::printNet() const{
  printf("Total number of neurons %d", nrOfNeurons);
  printf(" - Inputs : %d", nrOfInputs);
  printf(" - Hidden + Bias : %d" , (nrOfNeurons-nrOfInputs-nrOfOutputs));
  printf(" - Outputs : %d" , nrOfOutputs);
  if(actOrder!=NULL){
    for(int i=0; i < nrOfNeurons; i++){
      printf("Neuron %d priority %d links %d" , i , (actOrder[i]->getPriority()), (actOrder[i]->getNrOfLinks()) );
      for(int z=0; z < actOrder[i]->getNrOfLinks(); z++){
	printf(" %g" , ((double)actOrder[i]->getLinkWeight(z)/12.8));
      }
      printf(" type %d", actOrder[i]->getNodeType());
      printf(" activation %g" , actOrder[i]->getActivationValue());
    }
  }
}

/**
 *Mutate all neurons link weights in the network.
 *@param bitMutateProbability The chance that a bit in a weight gets flipped/mutated
 */
void Ann::mutateNeurons(double bitMutateProbability){
  for(int i=0; i < nrOfNeurons;i++){
    neurons[i]->mutateWeights(bitMutateProbability);
  }
}
/**
 *Mutate neurons link weights in the network where the neurons has the priority pri.
 *@param bitMutateProbability The chance that a bit in a weight gets flipped/mutated
 *@param pri The priority of the nodes to look for
 */
void Ann::mutateNeuronsWithPri(double bitMutateProbability, int pri){
  for(int i=0; i < nrOfNeurons;i++){
    neurons[i]->mutateWeightsWithPri(bitMutateProbability, pri);
  }
}

/**
 *Get the number of links belonging to a neuron.
 *@param index The internal index of the neuron.
 *@return The number of links or 0 if the neuron does not exist.
 */
int Ann::getNrOfLinks(int index) const{
  if(index < 0 || index > nrOfNeurons){
    return 0;
  }
  else{
    return actOrder[index]->getNrOfLinks();
  }
    
}
/**
 * Get the number of input neurons in the network.
 * @return The number of input neurons.
 */
int Ann::getNrOfInputNodes() const{
#ifdef ANN_DEBUG
  cout << "Ann::getNrOfInputNodes returns " << nrOfInputs << endl;
#endif
  return nrOfInputs;
}

/**
 *Get the number of neurons in a network.
 *@return The number of neurons.
 */
int Ann::getNrOfNodes() const{
#ifdef ANN_DEBUG
  cout << "Ann::getNrOfInputNodes returns " << nrOfNeurons << endl;
#endif
  return nrOfNeurons;
}
/**
 * Get the activation pointer from a neuron
 * @param index The internal index of the neuron
 * @return The pointer to the activation value or NULL if index is out of range.
 */
double *Ann::getActivationPointer(int index){
  if(index < 0 || index > nrOfNeurons){
    return NULL;
  }
  else{
    return neurons[index]->activationPointer();
  }
}
/**
 * Get the activation value from a neuron
 * @param index The internal index of the neuron
 * @return The activation value or 0.0 if index is out of range.
 */
double Ann::getActivationValue(int index) const{
  if(index < 0 || index > nrOfNeurons){
    return(0.0);
  }
  else{
    return neurons[index]->getActivationValue();
  }
}

/**
 * Get the number of output neurons in the network.
 * @return The number of output neurons.
 */

int Ann::getNrOfOutputNodes() const{
#ifdef ANN_DEBUG
  cout << "Ann::getNrOfOutputNodes returns " << nrOfOutputs << endl;
#endif
return nrOfOutputs;
}

/**
 * Replace the weights with the weights from another ANN.
 * @attention The ANN must have the same architecture.
 * @param from The ANN to copy the weights from.
 */
void Ann::copyWeightsFrom(class Ann *from){
  for(int i=0; i < nrOfNeurons; i++){
    for(int u=0; u < actOrder[i]->getNrOfLinks(); u++){
      actOrder[i]->setLinkWeight(u,from->getLinkWeight(i,u));
    }
  }
}

/**
 * Get the weight of a neurons link.
 * @param neuron The internal index of the neuron
 * @param link The internal index of the link in the neuron
 */
Weight Ann::getLinkWeight(int neuron,int link) const{
  return(actOrder[neuron]->getLinkWeight(link));
}
/**
 * Get a pointer to the weight of a neurons link.
 * @param neuron The internal index of the neuron
 * @param link The internal index of the link in the neuron
 */
Weight *Ann::getLinkWeightPointer(int neuron,int link){
  return(actOrder[neuron]->getLinkWeightPointer(link));
}

/**
 * Make a exact copy of the ANN
 * @return A identical copy of the ANN.
 */
class Ann* Ann::duplicate() const{
  Ann *copy;
  Weight linkTo,SCNto;
  int i,l,q,z;
  /* Calls Ann::Ann(int dontCare) instead of Ann:Ann()  */
  /* which means that the net doesnt get a bias node by */
  /* default. */
  copy = new Ann(1);
  
  /* Copy all neurons */
  for(i=0; i < nrOfNeurons; i++){
    copy->addNeuron(neurons[i]->getPriority(),
		    neurons[i]->getNodeType(),
		    neurons[i]->getActFunc(),
		    neurons[i]->getId());
  }

  for(i=0; i < nrOfNeurons; i++){
    for(l=0; l < neurons[i]->getNrOfLinks();l++){
      linkTo=neurons[i]->getLinkToId(l);
      for(q=0; q < nrOfNeurons; q++){
	if(linkTo==copy->neurons[q]->getId()){
	  if(neurons[i]->getLinkType(l)==SCN_L){
	    SCNto=neurons[i]->getLinkSCNid(l);
	    for(z=0; copy->neurons[z]->getId()!=SCNto &&
		  z < nrOfNeurons; z++);
	    copy->neurons[i]->createSCNLink(copy->neurons[q],copy->neurons[z]);
	  }
	  else{
	    copy->neurons[i]->createLink(copy->neurons[q],neurons[i]->getLinkWeight(l),neurons[i]->getLinkType(l));
	  }
	  break;
	}
      }
    }
  }
  return copy;
}

/**
 * Compare the ANN with another ANN, prints the result to stdout
 * @attention This method doesn't do any real valid tests.
 * @param net The ANN to compare with
 */
void Ann::compareAnn(class Ann *net) const{
   int error=0;
  if(nrOfInputs!=net->getNrOfInputNodes()){
    printf("Ann::compareAnn In node count miss match\n");
    error++;
  }
  if(nrOfOutputs!=net->getNrOfOutputNodes()){
    printf("Ann::compareAnn  Out node count mismatch\n");
     error++;
  }
   if(error){
      printf("This net look like this\n"); 
      printNet();
      printf("Compared net looked like this\n"); 
      net->printNet();
   }
}

/**
 * Constructor - initiate a ANN and adds a bias node with priority -32768
 */
Ann::Ann(){
  nrOfNeurons=0;
  nrOfSCNs=0;
  nrOfInputs=0;
  nrOfOutputs=0;
  actOrder=NULL;
  inNodes=NULL;
  outNodes=NULL;
  SCNNodes=NULL;
  addNeuron(-32768,BIAS,NONE);
  neurons[0]->setActivationValue(1.0); /* Bias Activationvalue allways 1.0 */
}
/**
 * Constructor - initiate a ANN, does not add a bias node.
 * @param dontCare A dummy variable.
 */
Ann::Ann(int dontCare){
  nrOfNeurons=0;
  nrOfInputs=0;
  nrOfOutputs=0;
  nrOfSCNs=0;
  actOrder=NULL;
  inNodes=NULL;
  outNodes=NULL;
  SCNNodes=NULL;
}

/**
 * Deconstructor, cleans up and deallocates all allocated memory.
 */

Ann::~Ann(){
  for(int i=0; i < nrOfNeurons; i++){
    delete(neurons[i]);
  }
  free(neurons);
  free(actOrder);
  free(inNodes);
  free(outNodes);
  free(SCNNodes);
}

/**
 * Adds a unique neuron to the network.
 * @param iLayerPriority The priority of the new neuron
 * @param inType The type of new neuron
 * @param iFuncT The activation function of the new neuron
 */
void Ann::addNeuron(int iLayerPriority, NODE_T inType, ACTIVATION_T iFuncT){
  if(nrOfNeurons==0){
    neurons = (class Neuron**) malloc(sizeof(class Neuron*));
    neurons[0]= new Neuron(iLayerPriority, inType, iFuncT);
    nrOfNeurons++;      
  }
  else{
    nrOfNeurons++;
    neurons = (class Neuron**) realloc(neurons,sizeof(class Neuron*)*nrOfNeurons);
    neurons[nrOfNeurons-1] = new Neuron(iLayerPriority, inType, iFuncT);
   }
  if(inType!=BIAS && inType!=INPUT){
    neurons[nrOfNeurons-1]->createLink(neurons[0],0,NORMAL);
  }
  if(inType==INPUT){
    addInNode(neurons[nrOfNeurons-1]);
  }
  if(inType==OUTPUT){
    addOutNode(neurons[nrOfNeurons-1]);
  }
  if(inType==SCN){
    addSCNNode(neurons[nrOfNeurons-1]);
  }
  sort();
}
/**
 * Adds a neuron to the network.
 * @attention Only use this method if you know what you are doing
 * @see Ann::addNeuron(int,NODE_T,ACTIVATION_T)
 * @param iLayerPriority The priority of the new neuron
 * @param inType The type of new neuron
 * @param iFuncT The activation function of the new neuron
 * @param if The unique id of the neuron
 */

void Ann::addNeuron(int iLayerPriority, NODE_T inType, ACTIVATION_T iFuncT, gint32 id){
  if(nrOfNeurons==0){
    neurons = (class Neuron**) malloc(sizeof(class Neuron*));
    neurons[0]= new Neuron(iLayerPriority, inType, iFuncT, id);
    nrOfNeurons++;      
  }
  else{
    nrOfNeurons++;
    neurons = (class Neuron**) realloc(neurons,sizeof(class Neuron*)*nrOfNeurons);
    neurons[nrOfNeurons-1] = new Neuron(iLayerPriority, inType, iFuncT, id);
   }
  if(inType==INPUT){
    addInNode(neurons[nrOfNeurons-1]);
  }
  if(inType==OUTPUT){
    addOutNode(neurons[nrOfNeurons-1]);
  }
  if(inType==SCN){
    addSCNNode(neurons[nrOfNeurons-1]);
  }
  sort();
}

/**
 * Compare the priority of two neurons. Used for various quicksorts.
 * If the priority is the same we compare them by their ID's since
 * qsort as an undefined behavour when items are equal, so
 * inorder to get the same behavour on different platforms we
 * make sure to use a unique identifier as the second sort key.
 * @param neuron1 Neuron A
 * @param neuron2 Neuron B
 * @return \f$ (A-B) \f$
 */
int comparePri(const void *neuron1, const void *neuron2){
  if((*(Neuron**)neuron1)->getPriority()==(*(Neuron**)neuron2)->getPriority())
    return((*(Neuron**)neuron1)->getId()==(*(Neuron**)neuron2)->getId());
  return((*(Neuron**)neuron1)->getPriority()
	 -(*(Neuron**)neuron2)->getPriority());
}

/**
 * Sort the internal neurons in activation order
 */
void Ann::sort(){
  //  if(NULL!=actOrder)
  free(actOrder);
  actOrder = (Neuron**)malloc(sizeof(Neuron*)*nrOfNeurons);
  for(int i=0; i < nrOfNeurons; i++)
    actOrder[i]=neurons[i];
  qsort(actOrder,nrOfNeurons,sizeof(Neuron*),comparePri);
}

/**
 * Set the activation of a input neuron.
 * @param index The index of the input neuron [0..nrOfInputs]
 * @param value The new activation value of the neuron
 */
void Ann::setInput(int index, double value){
  if(index < nrOfInputs)
    inNodes[index]->setActivationValue(value);
}
/**
 * Get the activation of a input neuron.
 * @param index The index of the input neuron [0..nrOfInputs]
 * @return The activation value of the input neuron or -1 if index is out of range.
 */
double Ann::getInput(int index) const{
  if(index < nrOfInputs)
    return inNodes[index]->getActivationValue();
  return(-1);
}
/**
 * Get the output of a output neuron.
 * @param index The index of the output neuron [0..nrOfOutputs]
 * @return The activation value of the output neuron or -1 if index is out of range.
 */

double Ann::getOutput(int index) const{
  if(index < nrOfOutputs)
    return outNodes[index]->getActivationValue();
  return -1;
}

/**
 * Get a activation pointer to a output neuron.
 * @param index The index of the output neuron [0..nrOfOutputs]
 * @return The activation value pointer or NULL if index is out of range.
 */
double *Ann::getOutputPtr(int index){
  if(index < nrOfOutputs)
    return outNodes[index]->activationPointer();
  return NULL;
}

/**
 * Add a help pointer to a input neuron
 * @param The newly created input neuron
 */
void Ann::addInNode(class Neuron *node){
  if(nrOfInputs==0){
    inNodes = (class Neuron**) malloc(sizeof(class Neuron*));
    inNodes[0] = node;
    nrOfInputs++;
  }else{
    inNodes = (class Neuron**) realloc(inNodes,sizeof(class Neuron*)*++nrOfInputs);
    inNodes[nrOfInputs-1]=node;    
  }
}

/**
 * Add a help pointer to a output neuron
 * @param The newly created output neuron
 */
void Ann::addOutNode(class Neuron *node){
  if(nrOfOutputs==0){
    outNodes = (class Neuron**) malloc(sizeof(class Neuron*));
    outNodes[0] = node;
    nrOfOutputs++;
  }else{
    outNodes = (class Neuron**) realloc(outNodes,sizeof(class Neuron*)*++nrOfOutputs);
    outNodes[nrOfOutputs-1]=node;    
  }
}
/**
 * Add a help pointer to a SCN neuron
 * @param The newly created SCN neuron
 */
void Ann::addSCNNode(class Neuron *node){
  if(nrOfSCNs==0){
    SCNNodes = (class Neuron**) malloc(sizeof(class Neuron*));
    SCNNodes[0] = node;
    nrOfSCNs++;
  }else{
    SCNNodes = (class Neuron**) realloc(SCNNodes,sizeof(class Neuron*)*++nrOfSCNs);
    SCNNodes[nrOfSCNs-1]=node;    
  }
}

/**
 * Removes a neuron from the ANN. Removes all links to the neuron and from the neuron.
 * @param The internal index of the neuron to remove.
 */
void Ann::removeNeuron(int index){
  int i;
  if(index==(nrOfNeurons-1)){
    for(i=0; i < nrOfNeurons; i++){
      neurons[i]->removeLinksTo(neurons[index]);
    }
    delete(neurons[index]);
    neurons = (class Neuron**) realloc(neurons,sizeof(class Neuron*)*--nrOfNeurons);
  }
  else{
    for(i=0; i < nrOfNeurons; i++){
      neurons[i]->removeLinksTo(neurons[index]);
    }
    delete(neurons[index]);
    for(i=index; i < (nrOfNeurons-1); i++){
      neurons[i]=neurons[i+1];
    }
    neurons = (class Neuron**) realloc(neurons,sizeof(class Neuron*)*--nrOfNeurons);
  }
}
/**
 * Create a link between two neurons with an inital weight
 * @param from The neuron where the links comes from
 * @param to The target of the link.
 * @param initWeight The inital weight of the link
 */
void Ann::connectNeurons(class Neuron *from, class Neuron *to,Weight initWeight){
  to->createLink(from,initWeight,NORMAL);
}
/**
 * Removes a link that goes from a neuron to a neuron.
 * @param from The neuron where the links comes from
 * @param to The target of the link.
 */
void Ann::removeConnection(class Neuron *from, class Neuron *to){
  to->removeLinksTo(from);
}

/**
 * Activate and propagate the network.
 *
 */
void Ann::activateNet(){
  for(int i=1; i < nrOfNeurons; i++){ /* First node is BIAS node */
    actOrder[i]->activate();
  }
}

/**
 * Reset all activation in the network to zero.
 * @attention The bias node (index 0) remains untouched
 *
 */
void Ann::resetNet(){
  for(int i=1; i < nrOfNeurons; i++){ /* First node is BIAS node */
    neurons[i]->setActivationValue(0.0);
  }
}

/**
 * Get a pointer to a input neurons activation value
 * @param index The internal input index of the neuron.
 */
double *Ann::getInputPointer(int index){
  return inNodes[index]->activationPointer();
}

/**
 * Connect two neuron layers (nodes with the same priority) using a third  layer whichs output should become the weights of the connection.
 * @attention Some sane tests will take place and exit will be called if they fail.
 * @param from The priority of the neurons where the links comes from
 * @param to The priority of the target layer for the links.
 * @param scPri The priority of the SCN layer which should be used.
 *
 */
void Ann::connectLayerWithSCN(int fromPri, int toPri, int scnPri){
  int SCNsRequired;
  int freeSCNs;
  int scnI;
  SCNsRequired = countNodesWithPri(toPri) *
    countNodesWithPri(fromPri);
  if(SCNsRequired==0){
    fprintf(stderr," Ann::connectLayerWithSCN there are no nodes with priority specified with fromPri and/or toPri\n");
    exit(-1);
  }
  freeSCNs = countUnUsedSCNNodesWithPri(scnPri);
  if(SCNsRequired > freeSCNs){
    fprintf(stderr," Ann::connectLayerWithSCN You have not created enough SCN neurons\n");
    exit(-1);
  }
  // Everything is ok - time to connect SCN nodes with links
  for(int i=0; i <nrOfNeurons; i++){
    if(neurons[i]->getPriority() == fromPri){
      for(int q=0; q < nrOfNeurons; q++){
	if(neurons[q]->getPriority() == toPri){
	  scnI = nextUnUsedSCNNodeWithPri(scnPri);
	  if(scnI>= 0){
	    neurons[q]->createSCNLink(neurons[i],SCNNodes[scnI]);
	  }
	  else{
	    fprintf(stderr," Ann::connectLayerWithSCN This error shouldnt occur\n");
	    fprintf(stderr," Ann::connectLayerWithSCN Something is wrong with the SCN code\n");
	    exit(-1);
	  }
	}
      }
    }
  }
}
/**
 * Connect/link one priority layer with another.
 * @param fromPri The priority of the from layer.
 * @param toPri The priority of the layer which the links goes to.
 */

void Ann::connectLayer(int fromPri, int toPri){
  for(int i=0; i <nrOfNeurons; i++){
    if(neurons[i]->getPriority() == fromPri){
      for(int q=0; q < nrOfNeurons; q++){
	if(neurons[q]->getPriority() == toPri)
	  connectNeurons(neurons[i],neurons[q],0);
      }
    }
  }
}
/**
 * Get the number of neurons of a given priority.
 * @param cntPri The priority to look for.
 * @return The number of neurons found with the given priority.
 */
int Ann::countNodesWithPri(int cntPri) const{
  int count = 0;
  for(int i=0; i < nrOfNeurons; i++){
    if(neurons[i]->getPriority()==cntPri)
      count++;
  }
  return count;
}
/**
 * Get the number of SCN neurons of a given priority.
 * @param scnPri The priority to look for.
 * @return The number of SCN neurons found with the given priority.
 */
int Ann::countSCNNodesWithPri(int scnPri) const{
  int count = 0;
  for(int i=0; i < nrOfSCNs; i++){
    if(SCNNodes[i]->getPriority()==scnPri)
      count++;
  }
  return count;
}

/**
 * Get the number of unused SCN neurons of a given priority.
 * That is SCN neurons not connected to a weight.
 * @param scnPri The priority to look for.
 * @return The number of unused SCN neurons found with the given priority.
 */
int Ann::countUnUsedSCNNodesWithPri(int scnPri) const{
  int count = 0;
  for(int i=0; i < nrOfSCNs; i++){
    if(SCNNodes[i]->getPriority()==scnPri &&
       SCNNodes[i]->SCNNotUsed())
      count++;
  }
  return count;
}

/**
 * Get the index of the next unused SCN neuron with a given priority.
 * @param scnPri The priority to look for.
 * @return The index of the neuron or -1 if no such neuron was found.
 */
int Ann::nextUnUsedSCNNodeWithPri(int scnPri) const{
  int index=0;
  for(index = 0; index < nrOfSCNs; index++){
    if(SCNNodes[index]->getPriority()==scnPri &&
       SCNNodes[index]->SCNNotUsed())
      break;
  }
  if(nrOfSCNs>0)
    return index;
  else
    return -1;
}
