/*
  YAKS, a Khepera simulator including a separate GA and ANN 
  (Genetic Algoritm, Artificial Neural Net).
  Copyright (C) 2000  Johan Carlsson (johanc@ida.his.se)
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 */

/* Standard libraries */
#include <stdio.h>
#include <unistd.h>
#include <values.h>
#include <stdio.h>
#include <fcntl.h>
#include <termio.h>
#include <time.h>

/* Simulator related includes*/ 
#include "geom.h"
#include "environ.h"
#include "param.h"
#include "real.h"
#include "terminal.h"

/* GUI related includes*/
#ifdef GUI
#include "monitor.h"
#include <pthread.h>
#include <semaphore.h>
#include <support.h>
#include <glib.h>
#include "world.h"

pthread_t tid;
// GUI requires this
int gen=0;

sem_t gui_run;
sem_t gui_step;
sem_t gui_ready;

                 
#endif /* GUI */

/* The world and such alike */
Environ *world;
Real *real;

#ifdef GUI
/* Callback functions for GUI */
void sim_redrawWorld(){
  printf("No - that would cause deadlock =) \n");
  //  world->redrawWorld();
}

float sim_getRobotX(){
  return world->robot[0]->getX();
}

float sim_getRobotY(){
  return world->robot[0]->getY();
}

void sim_setRobotX(float x){
  world->robot[0]->setX(x);
}
void sim_setRobotY(float y){
  world->robot[0]->setY(y);
}

void sim_setWorldSize(){
  gui_setWorldSize((int)(world->dx*xScale),
		   (int)(world->dy*xScale));
}
void sim_loadNextIndivid(){
}
#endif /* GUI */


void sim_shutdown(){
  exit(0);
}

void sigCtrlC(int dontCare){
  Terminal *term;
  term = new Terminal();
  term->addItem("q","Simulator shutdown",sim_shutdown);
  if(useSerialLine){
    for(int i=0; i < paramNrOfRobots; i++){
      /* Stop the robots */
      fprintf(stderr,"Telling bot %d to stop\n",i);
      real->runStep(0.5,0.5,i);
    }
  }
  term->runTerminal(); 
  delete(term);
}

#define T 0.3
void getMotorSimple(int i,float *l, float *r){
  if(input[i][2] > T || input[i][3] > T){
      *l=0.1; *r=0.9;
  }else if(input[i][5] > T){
    *l=0.6; *r=0.8;
  }else if(input[i][0] > T){
    *l=0.8; *r=0.6;
  }else{
    *l=0.75; *r=0.75;
  }
}

void getMotorCorridorFollower(int i, float *l, float *r){
  *l = 0.6 + (input[i][0] - input[i][5]);
  *r = 0.6 + (input[i][5] - input[i][0]);
}

void getMotorFollowWall(int i, float *l, float *r){
  if (input[i][4]*1024.0 > 200){
    /* turn left, in corner */
    *l = 0.3;
    *r = 0.75;
  }
  else if (input[i][5]*1024.0 < 500) {
    /* adjust right, losing wall */
    *l = 0.75;
    *r = 0.6;
  }
  else {
    /* ahead */
    *l = *r =0.75;
  }
}


int main(int argc,char **argv){

  char worldFileStr[200];
  int run;

  seed=time(NULL);
  srand(seed);
  setlinebuf(stdout);
  setlinebuf(stderr);

  if(argc > 1)
    loadParams(argv[1]);
  else{
    fprintf(stderr,"No option file specified\n");
    fprintf(stderr,"use: sim optionFile.opt \n\n");
    return -1;
  }

  /* Prepare the world */
  /* First we create the world */
  world = new Environ();
  /* We tell the world to load the world specified in the option file 
   * and all necessary sensor sample files will be loaded to 
   */
  sprintf(worldFileStr,"%s/%s",worldPath,worldFile);
  world->loadWorld(worldFileStr);
  /* Add all robots */
  for(int x=0; x < paramNrOfRobots; x++){
    if(robotStr[x]){
      world->addRobot(robotStr[x]);
    }
  }

  /* Our ctrl^c handler that starts a terminal */
  signal(SIGINT,sigCtrlC);  
#ifdef GUI

  g_thread_init(NULL);

  gtk_set_locale ();
  gtk_init (&argc, &argv);
  //  gdk_init(&argc, &argv);

  add_pixmap_directory ((const gchar*)"./pixmaps");
  add_pixmap_directory ((const gchar*)"./pixmaps");

  sim_setWorldSize();

  sem_init(&gui_ready,1,0); 
  sem_init(&gui_run,1,0);
  sem_init(&gui_step,1,1);

  pthread_create(&tid, NULL,(void *(*)(void *))gui_init,NULL);

  printf("Wait for gui to launch\n");
  sem_wait(&gui_ready);
  printf("GUI ready -> GO !!! \n");
  printf("Waiting for user to press run\n");

  sem_wait(&gui_run);

#endif /* GUI */     

  /*############################################################REAL###############*/
  if(useSerialLine){ /* Real environment */
    real = new Real();
    if(!real->init()){
      fprintf(stderr,"main() init of serial line failed\n");
      exit(-1);
    }
    world->resetEnvironment();
    /**************************************************/
    /*   Let the robot interact with the environment  */
    /*   for nrOfSteps time steps                     */
    for(;;){
      
      /**************************************************/
      /* Let the world generate sensor values for       */
      /* robot with index 0 -> create inputs values to  */
      /* ANN   */
      for(int i=0; i < paramNrOfRobots; i++){
	float l,r;
	real->setInput(i); 	 
	usleep(5000);	    
	getMotorFollowWall(i,&l,&r);
	real->runStep(l,r,i);
      }
      world->drawLayout();
      usleep(5000);
    }
  }
  /*############################################################SIMULATED##########*/
  else{/* Simulated environment*/
	  
    /* Reset robots, new startpos and such thingis */
    /* make sure  no objects are taken anymore  */
    /* if GUI redraw world */
    world->resetEnvironment();

    /**************************************************/
    /* Let the world generate sensor values for the   */
    /* robots                                         */
    for(int i=0; i < paramNrOfRobots; i++)
      world->setInput(i); 	 

    /**************************************************/
    /*   Let the robot interact with the environment  */
    /*   for nrOfSteps time steps                     */
    
    for(;;){
      for(int i=0; i < paramNrOfRobots; i++){
	float l,r;
	getMotorSimple(i,&l,&r);
	world->runStep(l,r,i);
	world->runGripper(0.5,0.5,i);
      }
#ifdef GUI
      if(gui_stepping)
	sem_wait(&gui_step);
#endif /* GUI */
      for(int i=0; i < paramNrOfRobots; i++){
	if(world->checkCrash(i)){
	  world->robot[i]->setX(world->dx/2);
	  world->robot[i]->setY(world->dy/2);
	}
	world->setInput(i); 	 	      
      }
#ifdef GUI	     
      world->drawBots();
#endif /* GUI */
    }
  }
#ifdef GUI
  pthread_join(tid,NULL);
#endif /* GUI */
}











