/*
  YAKS, a Khepera simulator including a separate GA and ANN 
  (Genetic Algoritm, Artificial Neural Net).
  Copyright (C) 2000  Johan Carlsson (johanc@ida.his.se)
  
  This program is g_free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the G_Free Software Foundation; either version 2
  of the License, or any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the G_Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 */

#include "environ.h"
#include <glib.h>
#include <string.h>

/**
 * @brief Constructor for a Environ (environment). 
 * Initiates a new world with \a paramNrOfRobots which is set by
 * a call to loadParams(). It also initiates the geometric library
 * by calling gemometricInit(). At last the newly created robots
 * are initiated by calling Robot::init().
 * @attention loadParams() must be called before creating a new Environ.
 * @see loadParams()
 * @see geometricInit()
 * @see Robot::init()
 */
Environ::Environ(){

  nrOfLights = 0;
  nrOfWalls = 0;
  nrOfObst = 0;
  nrOfSObst = 0;
  nrOfRobots = 0;
  nrOfZones = 0;
  wall = NULL;
  robot = NULL;
  start = NULL;
  obst = NULL;
  sobst = NULL;
  zone = NULL;
  light = NULL;
  dx = 0;
  dy = 0;
  geometricInit();
  robot =  new Robot*[paramNrOfRobots];
  bot_stat = (RobotDescription**)g_malloc(sizeof(RobotDescription*)*paramNrOfRobots);
}

/**
 * Add a robot to the world.
 * @param definition The text string describing the robot.
 */
void Environ::addRobot(const char *definition){
  char   tmpStr[300];
  int i=0;
  RobotDescription *bot;
  if(nrOfRobots < paramNrOfRobots){
    bot = (RobotDescription*)g_malloc(sizeof(RobotDescription));
    if(decodeRobot(definition,bot)){
      if(bot->useGripper){
	robot[nrOfRobots] =  new GripperRobot();
      }else{
	robot[nrOfRobots] = new Robot();
      }
      if(bot->hasRod){
	for(i=0; i < bot->rods; i++)
	  robot[nrOfRobots]->addRod(bot->rod_coords[i*2],bot->rod_coords[i*2+1]);
      }
      bot_stat[nrOfRobots] = bot;
      if(verboseLevel>0)
	printRobotInfo(bot);
      if(nrOfRobots==0){
	g_snprintf(tmpStr,300,"%s/%s",cameraPath,robotCamera);
	robot[0]->init(tmpStr);
      }
      nrOfRobots++;
    }else{
      g_free(bot);
      fprintf(stderr,"Environ::addRobot() Parsing of robot description failed\n");
      exit(EXIT_FAILURE);
    }
  }
}

/**
 * Decodes a string describing a robot.
 * @param def The string describing the robot.
 * @param b A pointer to a allocated robot description.
 * @return 1 if succesful otherwise 0.
 */
int Environ::decodeRobot(const char *def, RobotDescription *b){
  char *eater;
  int flags=0;
  int success=0;
  int l=0;
  unsigned int decimal=0;
  if(verboseLevel>1)
    printf("Robot description is %s\n",def);
  eater = g_strdup(def);
  success = sscanf(def," %i ",&flags);
  if(success != 1){
    success = sscanf(def," [%s] ",eater);
    if(success != 1){
      success = sscanf(def," b%s ",eater);
      if(success != 1){
	b->decodeMethod=BOT_FAILURE;
	return(0);
      }else{
	/* binary */
	if(verboseLevel>1)
	  printf("Binary mode %s\n",eater);
	b->decodeMethod=BOT_BINARY_CODED;
	l = strlen(eater);
	l--;
	for(int i=0; i < (l+1) && i < (int)sizeof(int)*8; i++){
	  switch(eater[i]){
	  case '0':
	    flags &= (~(1 << (l-i)));
	    decimal &= ~(1 << (l-i));
	    break;
	  case '1':
	    flags |= (1 << (l-i));
	    decimal |= (1 << (l-i));
	    break;
	  default:
	    i = l+1;
	    break;
	  }
	}

      }
    }else{
      /* text coded */
      b->decodeMethod=BOT_TEXT_CODED;
      flags=parseTextFlags(eater);
    }
  }else{
    b->decodeMethod=BOT_DECIMAL_CODED;
  }
  if(verboseLevel>1)
    printf(" => %0u => %#.4x \n",flags,flags);
  parseFlags(flags,b);
  parseRodCoords(def,b);
  g_free(eater);

  if(b->decodeMethod)
    return(1);
  else
    return(0);
}

/**
 * Extracts the coordinates of rods from a string.
 * @param The string.
 * @param A pointer to a allocated robot description struct.
 */
void Environ::parseRodCoords(const char *str,RobotDescription *b){
  char *s = g_strdup(str);
  char *p = NULL;
  unsigned int i;
  for(i=0; i < strlen(s); i++)
    if(s[i]=='#')
      s[i] = '\0';

  int paranb = countChar(s,'(');
  int parane = countChar(s,')'); 
  if(paranb != parane)
    fprintf(stderr,"Environ::parseRodCoords() Paranthesis missmatch in robot string: %s\n",str);
  if(b){
    b->hasRod=1;
    b->rods = paranb;
    p = s;
    for(i=0; i < strlen(s) && s[i] != '('; i++,p++);
    b->rod_coords = new float[b->rods*2];
    int w=0;
    for(i=0; i <(unsigned int) b->rods && w < b->rods*2; i++){
      if(sscanf(p," (%f,%f)",&b->rod_coords[w],&b->rod_coords[w+1])!=2)
	fprintf(stderr,"Environ::parseRodCoords() Parsing failed for rod coordinates: %s\n",p);
      w += 2;
      p++;
      while(p[0] != '(' && p[0]!='\0' && p++);
    }
  }
  g_free(s);
}

/**
 * Sets up a robot description using a flag variable.
 * @param flags The variable with the information.
 * @param b The allocated robot description struct.
 */
void Environ::parseFlags(int flags,RobotDescription *b){
  if(b){
    b->nrOfSensors=0;
    if(test(flags,BOT_2_INFRARED)){
      b->frontInfrared = 2;
      b->nrOfSensors += 2;
    }
    if(test(flags,BOT_4_INFRARED)){
      b->frontInfrared = 4;
      b->nrOfSensors += 4;
    }
    if(test(flags,BOT_6_INFRARED)){
      b->frontInfrared = 6;
      b->nrOfSensors += 6;
    }
    if(test(flags,BOT_BACK_INFRARED)){
      b->backInfrared = 1;
      b->nrOfSensors += 2;
    }
    else{
      b->backInfrared = 0;
    }
    if(test(flags,BOT_COMPASS_SENSOR)){
      b->useCompass = 1;
      b->nrOfSensors++;
    }
    else{
      b->useCompass = 0;
    }
    if(test(flags,BOT_ENERGY_SENSOR)){
      b->useEnergy = 1;
      b->nrOfSensors++;
    }
    else{
      b->useEnergy = 0;
    }
    if(test(flags,BOT_GRIPPER)){
      b->useGripper = 1;
    }
    else{
      b->useGripper = 0;
    }
    if(test(flags,BOT_GRIPPER_SENSOR)){
      b->useGripperSensor = 1;
      b->nrOfSensors++;
    }
    else{
      b->useGripperSensor = 0;
    }
    if(test(flags,BOT_GROUND_SENSOR)){
      b->useGround = 1;
      b->nrOfSensors++;
    }
    else{
      b->useGround = 0;
    }
    if(test(flags,BOT_LIGHT_SENSOR)){
      b->useLight = 1;
      b->nrOfSensors += 8;
    }
    else{
      b->useLight = 0;
    }
    if(test(flags,BOT_ROD_SENSOR)){
      b->useRodSensor = 1;
      b->nrOfSensors += ROD_RESOLUTION;
    }
    else{
      b->useRodSensor = 0;
    }
    if(test(flags,BOT_HAS_ROD))
      b->hasRod = 1;
    else
      b->hasRod = 0;
    
  }
}

int Environ::parseTextFlags(const char *eater){
  int flags=0;

  for(unsigned int i=0; i < strlen(eater); i++){
    switch(eater[i]){
    case 'B':
    case 'b':
      set(flags,BOT_BACK_INFRARED);
      break;
    case 'g':
    case 'G':
      set(flags,BOT_GRIPPER);
      break;
    case 'H':
    case 'h':
      set(flags,BOT_GRIPPER_SENSOR);
      break;
    case 'C':
    case 'c':
      set(flags,BOT_COMPASS_SENSOR);
      break;
    case 'R':
    case 'r':
      set(flags,BOT_ROD_SENSOR);
      break;
    case 'L':
    case 'l':
      set(flags,BOT_LIGHT_SENSOR);
      break;
    case 'E':
    case 'e':
      set(flags,BOT_ENERGY_SENSOR);
      break;
    case 'Z':
    case 'z':
      set(flags,BOT_GROUND_SENSOR);
      break;
    case 'P':
    case 'p':
      set(flags,BOT_HAS_ROD);
      break;
    case 'F':
    case 'f':
      i++;
      switch(eater[i]){
      case '2':
	set(flags,BOT_2_INFRARED);
	break;
      case '4':
	set(flags,BOT_4_INFRARED);
	break;
      case '6':
	set(flags,BOT_6_INFRARED);
	break;
      default:
	printf("Environ::parseTextFlags() \n");
	printf("---------------------------------------------------------\n");
	printf("Robot option F?, invalid variable '%c' defaults to 6 instead\n",eater[i]);
	printf("Valid values is [F6,F4,F2]\n");
	printf("---------------------------------------------------------\n");
	set(flags,BOT_6_INFRARED);
	i--;
	break;
      }
      break;
    default:	
      // silent ignore
      break;
    }
  }
  return(flags);
}
/**
 * Prints the information about a robot.
 * @param b Robot description describing the robot.
 */
void Environ::printRobotInfo(const RobotDescription *b) const{
  printf("___________________________________________________\n");
  printf("Robot - information\n");
  if(b->useGripper){
    printf("The robot have a gripper");
    if(b->useGripperSensor)
      printf(" and uses it's gripper sensor\n");
    else
      printf(".\n");
  }
  printf("Uses %d front infrared sensors.\n",b->frontInfrared);
  if(b->backInfrared)
    printf("Uses both back infrared.\n");
  if(b->useLight)
    printf("Uses light sensors.\n");
  if(b->useEnergy)
    printf("Uses the energy sensor.\n");
  if(b->useCompass)
    printf("Uses the compass sensor.\n");
  if(b->useGround)
    printf("Uses the bogus ground sensor.\n");
  if(b->useRodSensor)
    printf("Uses the rod camera.\n");
  if(b->hasRod){
    printf("Has %d rods :\n",b->rods);
    int w=0;
    for(int i=0; i < b->rods; i++){
      printf(" rod %d - %f, %f \n",i,
	     b->rod_coords[w],
	     b->rod_coords[w+1]);
      w += 2;
    }
  }
  
  printf("___________________________________________________\n");
}

/**
 * Resets all robots in the world.
 * @see Robot::resetRobot()
 */
void Environ::initWorld(){
  for(int i=0; i < nrOfRobots; i++)
    robot[i]->resetRobot();
}

/**
 *Deconstructor, deallocates all memory allocated.
 *
 */
Environ::~Environ(){
  int i;
  if(nrOfLights > 0){
    for(i=0; i < nrOfLights; i++){
      delete(light[i]);
    }
    g_free(light);
  }
  if(nrOfWalls > 0){
    for(i=0;i < nrOfWalls; i++){
      delete(wall[i]);
    }
    g_free(wall);
  }
  if(nrOfObst>0){
    for(i=0; i < nrOfObst; i++){
      delete(obst[i]);
    }
    g_free(obst);
  }

  if(nrOfSObst > 0){
    for(i=0; i < nrOfSObst; i++){
      delete(sobst[i]);
    }
    g_free(sobst);
  }
  if(nrOfRobots > 0){
    for(i=0; i < nrOfRobots; i++){
      delete(robot[i]);
    }
    g_free(robot);
  }
  if( nrOfZones > 0){
    for(i=0; i < nrOfZones; i++){
      delete(zone[i]);
    }
    g_free(zone);
  }
  
}

/*
 * Run a robot in the environment one timestep (100 ms)
 * @param o1 First argument for Robot::run() 
 * @param o2 Second argument for Robot::run()
 * @param robotNr Which robot to run.
 */

void Environ::runStep(float o1, float o2, int robotNr){
  robot[robotNr]->run(o1,o2);
}

/**
 * Get the X coordinate of Zone \a index.
 * @param index the internal index of the zone.
 * @return The x coordinate of the zone.
 */
float Environ::getZoneX(int index) const{
  return(zone[index]->x);
}
/**
 * Get the y coordinate of Zone \a index.
 * @param index the internal index of the zone.
 * @return The y coordinate of the zone.
 */
float Environ::getZoneY(int index) const{
  return(zone[index]->y);
}

/**
 * Set taken attribute on small round obstacle sobst::taken
 * @param index The internal index of the obstacle.
 */
void Environ::setTaken(int index){
  sobst[index]->taken=1.0;
}


#ifdef GUI
/**
 *  Draw sensors and world layout only.
 *
 */

void Environ::drawLayout(){
  if(gfxUpdate){
#ifndef OPENGL
    gui_drawBG();
#endif
    for(int robotNr=0; robotNr < nrOfRobots; robotNr++)
      gui_updateSensorWindow(robotNr,bot_stat[robotNr]->useRodSensor,bot_stat[robotNr]->useEnergy);
    gui_redrawWorld();
  }
}
#ifdef OPENGL
void Environ::drawBots(){
  drawLayout();
}
#else

/**
 * Draw all robots to the screen
 * @attention This method is only available when \a GUI is defined
 */
void Environ::drawBots(){
  Robot *rob;
  GripperRobot *grob;
  int ldir,rdir;
  int a[10];
  int b[10];
  
  if(gfxUpdate){
    gui_drawBG();
    for(int robotNr=0; robotNr < nrOfRobots; robotNr++){
      
      rob = robot[robotNr];

      gui_updateSensorWindow(robotNr,bot_stat[robotNr]->useRodSensor,bot_stat[robotNr]->useEnergy);      
      if(notTrajactoryRobot){
	gui_drawRobot((int)(xScale*rob->getX()),
		      (int)(xScale*rob->getY()),
		      rob->direction,
		      (int)(xScale*rob->getRadius()),
		      1,robotNr,bot_stat[robotNr]->useRodSensor);
	if(rob->isaGripperRobot()){
	  grob = (GripperRobot*) rob;
	  for(int i=0; i < 4; i++){
	    a[i] = (int)(xScale*grob->lgripper[i]);
	    b[i] = (int)(xScale*grob->rgripper[i]);
	  }
	  
	  ldir = g_controlAngle((int)(rob->direction + 90));
	  rdir = g_controlAngle((int)(rob->direction - 90));
	  b[4] = (int)(xScale * grob->flx);
	  b[5] = (int)(xScale * grob->fly);
	  b[6] = (int)(xScale * g_displaceX(grob->gx,
					    77.0/2.0,
					    ldir));
	  b[7] = (int)(xScale * g_displaceY(grob->gy,
					    77.0/2.0,
					    ldir));

	  a[4] = (int)(xScale * grob->frx);
	  a[5] = (int)(xScale * grob->fry);
	  a[6] = (int)(xScale * g_displaceX(grob->gx,
					    77.0/2.0,
					    rdir));
	  a[7] = (int)(xScale * g_displaceY(grob->gy,
					    77.0/2.0,
					    rdir));
	  
 	  gui_drawGripper(a,b,
			  1,
			  robotNr,
			  8
			  );
	}
      }
      else{
	gui_drawRobot((int)(xScale*rob->getX()),
		      (int)(xScale*rob->getY()),
		      rob->direction,
		      (int)(xScale*rob->getRadius()),
		      2,robotNr,bot_stat[robotNr]->useRodSensor);
	if(rob->isaGripperRobot()){
	  grob = (GripperRobot*) rob;
	  for(int i=0; i < 4; i++){
	    a[i] = (int)(xScale*grob->lgripper[i]);
	    b[i] = (int)(xScale*grob->rgripper[i]);
	  }
	  gui_drawGripper(a,b,
			  2,
			  robotNr,
			  2);
	}
      }
    }
    if(!notTrajactoryRobot)
      gui_drawBG();

    for(int i=0; i < nrOfSObst; i++){
       if(sobst[i]->taken==0.0)
	 gui_drawSObst((int)(xScale * sobst[i]->x),
		       (int)(xScale * sobst[i]->y),
		       (int)(xScale * sobst[i]->r)
		       );

    }
    gui_redrawWorld();
  }
}
#endif /* GUI */
#endif /* !OPENGL */
/**
 * Get the number of zones in the world.
 * @return The number of zones.
 */
int Environ::getNrOfZones() const{
  return(nrOfZones);
}
/**
 * Get the number of walls in the world.
 * @return The number of walls.
 */
int Environ::getNrOfWalls() const{
  return(nrOfWalls);
}
/**
 * Get the number of small round obstacles in the world.
 * @return The number of small round obstacles.
 */
int Environ::getNrOfSObst() const{
  return(nrOfSObst);
}
/**
 * Get the number of light sources in the world.
 * @return The number of light sources.
 */
int Environ::getNrOfLights() const{
  return(nrOfLights);
}

/**
 * Get the number of round obstacles in the world.
 * @return The number of round obstacles.
 */
int Environ::getNrOfObst() const {
  return(nrOfObst);
}

/**
 * Checks if robot \a rR is carrying something and if true moves that object according to the robots movement.
 * @param rR The robot.
 */
void Environ::moveSObst(int rR){
  int car;
  car = robot[rR]->carry; 
  if(car > -1)
    if(car > -1 && car < nrOfSObst)
      sobst[car]->tempMove(robot[rR]->getLastDist(),robot[rR]->getLastAngR());
}

/**
 * Simulates the gripper of a robot for 100 ms.
 * @param arm Arm position to reach [0.0, down back -  1.0, down front].
 * @param gripper Gripper position to reach [0.0 open, 1.0 close].
 * @param robotNr Which robots gripper..
 */
void Environ::runGripper(float arm, float gripper, int robotNr){
  class GripperRobot *rob;
  int obstPresent=0;
  int obstPosTaken=-1;
  int i;
  if(robot[robotNr]->isaGripperRobot()){
    rob = (class GripperRobot*)robot[robotNr];
    rob->setArmPosition(arm);
    rob->updateGripperCords();

    rob->clrGripperSensor();
    rob->clearCrashed();
    rob->updateThingInHandsCoord();
    if(rob->armDown()){
      for(i=0; i < nrOfWalls; i++){
	if(rob->checkCrash(wall[i]->x1,wall[i]->y1,wall[i]->x2,wall[i]->y2))
	  rob->armCrashed();
      
      
      }
      for(i=0; i < nrOfObst; i++){
	if(rob->checkCrash(obst[i]))
	  rob->armCrashed();
      }
      /*   check for object between hands */
      for(i=0; i < nrOfSObst; i++){
	/*   activate sensor if object present */
	if(sobst[i]->taken == 0.0){
	  switch(rob->inHands(sobst[i])){
	  case GE_OBJ_PRESENT:
	    obstPresent++;
	    obstPosTaken = i;
	    break;
	  case GE_OBJ_CRASH:
	    rob->armCrashed();
	    break;
	  default:
	    /* no operation */
	    break;
	  }
	}
      }
    }
    /* Run gripper hands and detect changes */
    if(rob->setGripperPosition(gripper)){
      if(rob->handsClosed()){
	if(obstPresent){
	  if(!rob->getCrashed())
	    rob->setThingInHands(sobst[obstPosTaken],
				 obstPosTaken);
	}
      }else{
	if(rob->holdingSomething()){
	  rob->releaseThingInHands();
	}
      }
    }
    rob->getSane();
  }
}

/**
 * Check if a coordinate is outside the world.
 * @attention This method makes the assumption that the world is box formed.
 * @param x The x coord.
 * @param y The y coord.
 * @return 1 if outside otherwise 0.
 */
int Environ::outside(float x, float y) const{
  if(x > dx || x < 0)
    return(1);
  if(y > dy || y < 0)
    return(1);
  return(0);
}

/**
 * Reset the environment, resets robots, marks objects not taken moves obstacled if specified.
 * @see resetRobot()
 * @see resetObstacle()
 */
void Environ::resetEnvironment(){
  int i;
  /* Relocate robots, random or fixed position */
  for(i=0; i < nrOfRobots; i++){
    resetRobot(i);
  }
  /* Mark all small round obstacles not taken */
  /* and move them if moveObstacles=1  */
  for(i=0; i < nrOfSObst; i++){
    resetObstacle(i); 
  }
#ifdef GUI
#ifndef OPENGL /* avoid flickering (gets redrawn later anyway) */
  if(gfxUpdate) /* show environment after reset */
    redrawWorld();
#endif
#endif /* GUI */
}

#ifdef GUI
#ifdef OPENGL
void Environ::redrawWorld(){
  gui_redrawWorld();
}
#else
/**
 * Draws the world to the gui back ground.
 */
void Environ::redrawWorld(){
  gui_clear();
  int i;
  for(i=0; i <nrOfWalls; i++){
    gui_drawWall((int)(xScale*wall[i]->x1),
	       (int)(xScale*wall[i]->y1),
	       (int)(xScale*wall[i]->x2),
	       (int)(xScale*wall[i]->y2));
  }
  for(i=0; i < nrOfSObst; i++){
    gui_drawSObst((int)(xScale*sobst[i]->x),
		(int)(xScale*sobst[i]->y),
		(int)(xScale*sobst[i]->r));
  }
  for(i=0; i < nrOfObst; i++){
    gui_drawObst((int)(xScale*obst[i]->x),
	       (int)(xScale*obst[i]->y),
	       (int)(xScale*obst[i]->r));
  }
  for(i=0; i < nrOfZones; i++){
    gui_drawZone((int)(xScale*zone[i]->x),(int)(xScale*zone[i]->y),(int)(xScale*zone[i]->r));
  }
  for(i=0; i < nrOfLights; i++){
    gui_drawLight((int)(xScale*light[i]->x),(int)(xScale*light[i]->y),(int)(xScale*light[i]->r));
  }
  gui_drawBG();
  gui_redrawWorld();
}
#endif /* !OPENGL */
#endif /* GUI */

/**
 * Check if robot \a robotNr is in zone \a index.
 * @param robotNr Which robot.
 * @param index Internal index of the zone.
 * @return Returns 1 if in a zone otherwise 0, -1 if indexes is out of range.
 */
int Environ::inZoneI(int robotNr, int index) const{
  float rx,ry;
  if(robotNr<paramNrOfRobots && index < nrOfZones){
    rx=robot[robotNr]->getX();
    ry=robot[robotNr]->getY();
    if(zone[index]->inZone(rx,ry,robotNr)){
      return 1;
    }
    return 0;
  }else{
    return -1;
  }
}

/**
 * Check if robot \a robotNr is in a zone.
 * @param robotNr Which robot.
 * @return Returns the zone index else -1.
 */
int Environ::inZone(int robotNr) const{
  float rx,ry;

  rx=robot[robotNr]->getX();
  ry=robot[robotNr]->getY();

  for(int i=0; i < nrOfZones; i++){
    if(zone[i]->inZone(rx,ry,robotNr)){
      return i;
    }
  }
  return -1;
}

/**
 * Mark all zones not visited.
 * @see Zone::clearVisit()
 */
void Environ::resetZones(){
  for(int i=0; i < nrOfZones; i++){
    zone[i]->clearVisit();
  }
}
/**
 * Get how many different zones the robot have visited.
 * @param robotNr Which robot.
 * @return The number of zones visited.
 */
int Environ::zonesVisited(int robotNr) const{
  int visits = 0;

  for(int i=0; i < nrOfZones; i++){
    if(zone[i]->visits(robotNr)){
      visits++;
    }
  }  
  return visits;
}

/**
 * Resets small round obstacle with index \a whichObstacle.
 * @param whichObstacle Index of obstacle.
 * @see sobst::resetMove
 * @see sobst::taken
 */
void Environ::resetObstacle(int whichObstacle){
  float x,y;
  // if been carried
  sobst[whichObstacle]->resetMove();
  sobst[whichObstacle]->taken=0; 
  if (moveObstacles){
    x = g_rans(dx);
    y = g_rans(dy);
    while (notPlacedOk(whichObstacle,x,y,sobst[whichObstacle]->r)){
      x = g_rans(dx);
      y = g_rans(dy);
    }	   
    sobst[whichObstacle]->x = x;
    sobst[whichObstacle]->y = y;
  }
}

/**
 * Check if a _small_ round obstacle is placed ok. (Not placed on an another object)
 * @param obstNr Index of the obstacle.
 * @param x X coordinate of obstacle.
 * @param y Y coordinate of obstacle.
 * @param radius There for historical reasons.
 * @return 1 if collision with an object is detected otherwise 0.
 */
int Environ::notPlacedOk(int obstNr, float x, float y, float radius) const{
  float tmp=0;
  float r;
  int i;
  r = sobst[obstNr]->r;
  for(i=0; i < nrOfSObst; i++){
    if(i!=obstNr){
      tmp = g_distPoint(x, y, sobst[i]->x, sobst[i]->y) - sobst[i]->r;
      if(tmp < r)
	return 1;
    }
  }
  for(i=0; i < nrOfObst; i++){
    tmp = g_distPoint(x, y, obst[i]->x, obst[i]->y) - obst[i]->r;
    if(tmp < r)
      return 1;
  }
  for(i=0; i < nrOfWalls; i++){
     tmp = g_distLinePoint(wall[i]->x1, wall[i]->y1,
		      wall[i]->x2, wall[i]->y2,
		      x,y);
    if(tmp < r)
      return 1;
  }
  return 0;
}
/**
 * Load world for the environment, and initates all obstacles.
 * @param fileName The world to load.
 */
void Environ::loadWorld(const char *fileName){

   FILE *fp;
   float  radius=0;
   float  sradius=0;
   char   tmpStr[300];
   char   buf[200];
   if(fileName==NULL){
     printf("Environ::loadWorld No world file specified in option file\n");
     exit(-1);
   }
   if (verboseLevel > 0)
     printf("Loading world %s\n",fileName);
   fp=fopen(fileName, "r");
   if(fp==NULL){
     printf("Couldn't open worldfile %s for reading",fileName);
     exit(-1);
   }
   nrOfWalls = nrOfZones = nrOfLights = nrOfObst = nrOfSObst = nrOfStartPositions = 0;

   while(!feof(fp)){
     fgets(buf,200,fp);
     switch(buf[0]){
     case 'w':
       wall = (class Wall**) g_realloc(wall,sizeof(class Wall*) * (nrOfWalls + 1));
       wall[nrOfWalls] = new Wall(buf);
       if (wall[nrOfWalls]->x1 > dx)
	 dx =(int) wall[nrOfWalls]->x1;
       if (wall[nrOfWalls]->x2 > dx)
	 dx = (int) wall[nrOfWalls]->x2;
       if (wall[nrOfWalls]->y1 > dy)
	 dy = (int) wall[nrOfWalls]->y1;
       if (wall[nrOfWalls]->y2> dy)
	 dy = (int)wall[nrOfWalls]->y2; 
       nrOfWalls++;
       break;
     case 'z':
       zone = (class Zone**) g_realloc(zone,sizeof(class Zone*) * (nrOfZones + 1));
       zone[nrOfZones] = new Zone(buf);
       nrOfZones++;
       break;
     case 'l':
       light = (class Light**) g_realloc(light,sizeof(class Light*) * (nrOfLights + 1));
       light[nrOfLights] = new Light(buf);
       nrOfLights++;
       break;
     case 'r':
       switch(buf[1]){
       case 'a':
	 sscanf(buf,"radius %f\n", &radius);
	 break;
       case 'o':
	 obst = (class Obst**) g_realloc(obst, sizeof(class Obst*) * (nrOfObst + 1));
	 obst[nrOfObst] = new Obst(buf);
	 nrOfObst++;
	 break;
       default:
	 printf("ignoring %s in worldfile %s\n",buf,fileName);
       }
       break;
     case 's':
       switch(buf[1]){
       case 'r':
	 switch(buf[2]){
	 case 'a':
	   sscanf(buf,"sradius %f\n", &sradius);
	   break;
	 case 'o':
	   sobst = (class Sobst**) g_realloc(sobst, sizeof(class Sobst*) * (nrOfSObst + 1));
	   sobst[nrOfSObst] = new Sobst(buf);
	   nrOfSObst++;
	   break;
	 default:
	   printf("ignoring %s in worldfile %s\n",buf,fileName);
	 }
	 break;
       case 't':
	 start = (float**) g_realloc(start,(nrOfStartPositions+1)*sizeof(float*));
	 start[nrOfStartPositions] = (float *) g_malloc(3 * sizeof(float));
	 sscanf(buf,"start %f %f %f",
		&start[nrOfStartPositions][0],
		&start[nrOfStartPositions][1],
		&start[nrOfStartPositions][2]);
	 nrOfStartPositions++;
	 break;
       case 'o':
	 sobst = (class Sobst**) g_realloc(sobst, sizeof(class Sobst*) * (nrOfSObst + 1));
	 sobst[nrOfSObst] = new Sobst(buf);
	 nrOfSObst++;
	 break;
       default:
	 printf("ignoring %s in worldfile %s\n",buf,fileName);
       }
       break;
     default:
       printf("ignoring %s in worldfile %s\n",buf,fileName);
     }
   }
   fclose(fp);

   /* init all */
   if(nrOfWalls>0){
     g_snprintf(tmpStr,300,"%s/%s",cameraPath,wallCamera);
     wall[0]->init(tmpStr);
   }
   if(nrOfObst > 0){
     obst[0]->setRadius(radius);   /* Common for all round obstacles (declared static in class Obst) */
     g_snprintf(tmpStr,300,"%s/%s",cameraPath,bigCamera);
     obst[0]->init(tmpStr);
   }
   if(nrOfSObst > 0){
     g_snprintf(tmpStr,300,"%s/%s",cameraPath,smallCamera);
     sobst[0]->init(tmpStr);
     sobst[0]->setRadius(sradius);
   }
   if(nrOfLights>0){
     g_snprintf(tmpStr,300,"%s/%s",cameraPath,lightCamera);     
     light[0]->init(tmpStr);
   }

#ifdef GUI
   gui_setWorldSize(dx, dy);
#endif /* GUI */
}

/**
 * Save the current world to file (here for the future - world editor).
 * @param worldFileName The name of file to save to.
 *
 */
void Environ::saveWorld(const char *worldFileName){
  int i,ii;
  FILE *fp;
  if(nrOfStartPositions > 0){
    for (i=0; i < nrOfStartPositions; i++){
      start[i][0] = (float) g_mrand(dx-40) + 20;	 
      start[i][1] = (float) g_mrand(dy-40) + 20;	 	   
      while (checkCrash(start[i][0],start[i][1]) != 0){
	start[i][0] = (float) g_mrand(dx-40) + 20;	
	start[i][1] = (float) g_mrand(dy-40) + 20;	 	 
      }	   
      start[i][2] = (float) g_mrand(360);	 	   	   
    } 	  
    fp=fopen(worldFileName, "w");
    if(fp==NULL){
      printf("Couldn't open world file %s for writing\n",worldFileName);
      exit(-1);
    }
    fprintf(fp,"%d %d %d %d %d %d\n",nrOfWalls ,nrOfZones,nrOfLights,nrOfObst,nrOfSObst,nrOfStartPositions);	 

    for (i=0; i < nrOfWalls; i++)
      fprintf(fp,"wall %f %f %f %f\n",
	      wall[i]->x1,wall[i]->y1,
	      wall[i]->x2,wall[i]->y2);

    for (i=0; i < nrOfZones; i++)
      fprintf(fp,"zone %f %f %f\n",
	      zone[i]->x,zone[i]->y,
	      zone[i]->r);

    for (i=0; i < nrOfLights; i++)
      fprintf(fp,"light %f %f\n", 
	      light[i]->x,light[i]->y);

    if (nrOfObst > 0){
      fprintf(fp,"radius %f\n", obst[0]->r);                  

      for (i=0; i < nrOfObst; i++)
	fprintf(fp,"roundobst %f %f\n", 
		obst[i]->x,obst[i]->y);
    }

    if (nrOfSObst > 0){
      fprintf(fp,"sradius %f\n", sobst[0]->r);                         

      for (i=0; i < nrOfSObst; i++)
	fprintf(fp,"sroundobst %f %f\n", 
		sobst[i]->x,sobst[i]->y);       
    }

    for (i=0; i < nrOfStartPositions; i++){
      for (ii=0; ii < 3; ii++)        
	fprintf(fp,"%f ", start[i][ii]);	 
      fprintf(fp,"\n");	     
    }
    fclose(fp);       
  }
}
/**
 * Print the current world to stdout.
 */
void Environ::printWorld() const{
  int i,ii;
  printf("#world size = %dx%d \n",dx,dy);   
  printf("#%d walls  :",nrOfWalls);        
  for (i=0; i < nrOfWalls; i++)
    printf("%.0f %.0f %.0f %.0f  ",
	   wall[i]->x1,
	   wall[i]->y1,
	   wall[i]->x2,
	   wall[i]->y2
	   );
  printf("\n");
  if (nrOfZones > 0){
    printf("#%d zones  :",nrOfZones);        
    for (i=0; i < nrOfZones; i++)
      printf("%.0f %.0f %.0f  ",
	     zone[i]->x,
	     zone[i]->y,
	     zone[i]->r
	     );
    printf("\n");
  }
  if (nrOfLights > 0){
    printf("#%d lightbulbs  :",nrOfLights);        	  
    for (i=0; i < nrOfLights; i++)
      printf("%.0f %.0f  ",
	     light[i]->x,
	     light[i]->y
	     );
    printf("\n");            
  }
  if (nrOfObst > 0){
    printf("#%d round obstacles :",nrOfObst);        	  	  
    printf("radius %.0f  ",obst[0]->r);          
    for (i=0; i < nrOfObst; i++)
      printf("%.0f %.0f  ", 
	     obst[i]->x,
	     obst[i]->y
	     );
    printf("\n");                 
  }
  if (nrOfSObst > 0){
    printf("#%d small round obstacles :",nrOfSObst);        	  	  
    printf("sradius %.0f  ", sobst[0]->r);          
    for (i=0; i < nrOfSObst; i++)
      printf("%.0f %.0f  ", 
	     sobst[i]->x,
	     sobst[i]->y
	     );
    printf("\n");                 
  }      
  if (nrOfStartPositions > 0){
    printf("starts posistions:\n");	     
    for (i=0; i < nrOfStartPositions; i++){
      printf("%d: ",i);
      for (ii=0; ii < 5; ii++)        
	printf("%.1f\t", start[i][ii]);	 
      printf("\n");	     
    } 
  }
}

/**
 * Method for placing all small round obstacle in a cluster.
 * @param cx Center X coordinate of the cluster.
 * @param cy Center Y coordinate of the cluster.
 * @param rd The distance to between the objects in the cluster.
 */
void Environ::placeSObstStack(float cx, float cy, float rd){
  int dimx = 0, dimy = 0;
  float r,sx,sy;
  int p,dx,dy;

  if(nrOfSObst>0){
    dimx = dimy = (int)sqrt((float)nrOfSObst);
    if(dimx*dimy != nrOfSObst)
      dimy++;

    r=sobst[0]->r;
    for(sx = cx-(r * dimx) - (rd*dimx)/2, p=0, dx=0; dx < dimx && p < nrOfSObst; dx++,sx+= (2 * r) + rd){
      for(sy = cy-(r * dimy) - (rd*dimy)/2 , dy=0; dy < dimy && p < nrOfSObst; dy++, sy+= (2 * r) + rd){
	sobst[p]->x = sx;
	sobst[p]->y = sy;
	p++;
      }
    }
  }else{
    fprintf(stderr,"Environ::placeSObstStack There were no small round obstacles to place in a stack\n");
  }
}

/**
 * Check if a robot has chrashed with something. 
 * @param whichRobot Which robot to check.
 * @param which Returns the index of the object which the robot crashed with or -1 if no crash.
 * @return The type of crash.
 * @see enum::crash_t
 */
crash_t Environ::checkCrash(int whichRobot, int *which) const{
  int i;
  double  dist;
  double  mindist;
  float   p1,p2;
  double direction;

  p1 = robot[whichRobot]->getX();
  p2 = robot[whichRobot]->getY();
  direction = robot[whichRobot]->direction;
  if(robot[whichRobot]->isaGripperRobot()){
    GripperRobot *rob;
    rob = (GripperRobot*)robot[whichRobot];
    if(rob->getCrashed()){
      *which = -1;
      return(GRIPPER_CRASH);
    }
  }
  mindist = robot[whichRobot]->getRadius();
  for(i=0;i<nrOfWalls;i++){
    dist = wall[i]->distR(p1,p2);
    if (dist < mindist){
      *which = i;
      return WALL_CRASH;
    }
  }
  for(i=0;i<nrOfObst;i++){
    dist = g_distPoint(p1,p2,obst[i]->x,obst[i]->y)-obst[0]->r;
    if (dist < mindist){
      *which = i;
      return OBST_CRASH;
    }
  }
  for(i=0; i < nrOfRobots; i++){
    dist = g_distPoint(p1, p2, robot[i]->getX(), robot[i]->getY()) - robot[i]->getRadius();
    if(dist < mindist && i!=whichRobot){
      *which = i;
      return ROBOT_CRASH;
    }
  }
  for(i=0;i<nrOfSObst;i++){
    if (sobst[i]->taken == 0.0){
      dist = g_distPoint(p1,p2,sobst[i]->x,sobst[i]->y)-sobst[0]->r;
      if (dist < mindist){
	*which = i;
	return SOBST_CRASH;
      }
    }
  }
 
  *which = -1;
  return(NO_CRASH);
}

/**
 * Check if a robot has chrashed with something.
 * @param p1 Robot X coordinate.
 * @param p2 Robot Y coordinate.
 * @return 1 if the robot has crashed with somethin otherwise 0.
 * @attention Not gripper safe.
 */
int Environ::checkCrash(float p1, float p2) const{
  int i;
  double  dist;
  double  mindist;
  
  mindist = robot[0]->getRadius();
  for(i=0;i<nrOfWalls;i++){
    dist = wall[i]->distR(p1,p2);
    if (dist < mindist)
      return 1;
  }
  for(i=0;i<nrOfObst;i++){
    dist = g_distPoint(p1,p2,obst[i]->x,obst[i]->y)-obst[0]->r;
    if (dist < mindist)
      return 1;
  }
  for(i=0;i<nrOfSObst;i++){
    if (sobst[i]->taken == 0.0){
      dist = g_distPoint(p1,p2,sobst[i]->x,sobst[i]->y)-sobst[0]->r;
      if (dist < mindist)
	return 1;
    }
  }
  for(i=0; i < nrOfRobots; i++){
    dist = g_distPoint(p1,p2,robot[i]->getX(),robot[i]->getY())-robot[i]->getRadius();
    if (dist < mindist)
      return 1;
  }
  return(0);
}


/**
 * Check if a robot has chrashed with something.
 * @param whichRobot Which robot.
 * @return 1 if the robot has crashed with somethin otherwise 0.
 */
int Environ::checkCrash(int whichRobot) const{
  int i;
  double  dist;
  double  mindist;
  float   p1,p2;
  double direction;

  p1 = robot[whichRobot]->getX();
  p2 = robot[whichRobot]->getY();
  direction = robot[whichRobot]->direction;
  if(robot[whichRobot]->isaGripperRobot()){
    GripperRobot *rob;
    rob = (GripperRobot*)robot[whichRobot];
    if(rob->getCrashed()){
      return(1);
    }
  }
  mindist = robot[whichRobot]->getRadius();
  for(i=0;i<nrOfWalls;i++){
    dist = wall[i]->distR(p1,p2);
    if (dist < mindist)
      return 1;
  }
  for(i=0;i<nrOfObst;i++){
    dist = g_distPoint(p1,p2,obst[i]->x,obst[i]->y)-obst[0]->r;
    if (dist < mindist)
      return 1;
  }
  for(i=0;i<nrOfSObst;i++){
    if (sobst[i]->taken == 0.0){
      dist = g_distPoint(p1,p2,sobst[i]->x,sobst[i]->y)-sobst[0]->r;
      if (dist < mindist)
	return 1;
    }
  }
  for(i=0; i < nrOfRobots; i++){
    dist = g_distPoint(p1, p2, robot[i]->getX(), robot[i]->getY()) - robot[i]->getRadius();
    if(dist < mindist && i!=whichRobot)
      return 1;
  }
  
 
  return(0);
}

/**
 * Reset khepera position. Random new position or according to start position. Clears gripper crash status.
 * @param robotNr Which robot to reset.
 * @see loadParams()
 */
void Environ::resetRobot(int robotNr){

  if (testSameStartPosition > 0){
    if(nrOfStartPositions > 0 && nrOfStartPositions > robotNr){
      robot[robotNr]->setX(start[robotNr][0]);
      robot[robotNr]->setY(start[robotNr][1]);
      robot[robotNr]->direction = start[robotNr][2];
    }
    else{
      fprintf(stderr,"You havent specified enough startpositions for\
testSameStartPositions in your world file\n");
      exit(-1);
    }
  }
  else{
    /* Make sure so that the gripper isn't causing us any 
     * trouble.
     **/ 
    if(robot[robotNr]->isaGripperRobot()){
      GripperRobot *bot;
      bot = (GripperRobot*)robot[robotNr];
      bot->clearCrashed();
    }
    robot[robotNr]->setX(g_rans(dx));
    robot[robotNr]->setY(g_rans(dy));
    robot[robotNr]->direction = fabs(g_rans(360.0));   
    while (checkCrash(robotNr) == 1 ){
      robot[robotNr]->setX(g_rans(dx));
      robot[robotNr]->setY(g_rans(dy));
    }	   
  }
}
/**
 * Prints the input array of a robot and a user added message to stdout.
 * @param robotNr Input for which robot.
 * @param str User string to print.
 */
void Environ::printInput(int robotNr, const char *str) const{
  int i;

  printf("----ROBOT %d input; START -> %s\n",robotNr,str);
  for(i=0; i < NETWORK_INPUTS; i++)
    printf("%.2f ",input[robotNr][i]);
  printf("\n");
  printf("----ROBOT %d input; END -> %s\n",robotNr,str);
}

/**
 * Adds the inputs of \a sens to a robots infrared sensors. All according to nrOfFrontInfrared.
 * @param robotNr For which robot.
 * @param sens Array of 8 floats.
 */
void Environ::addToInputArray(int robotNr, float *sens){
  int ii=0;
  switch(bot_stat[robotNr]->frontInfrared){
  case 6:
    for(ii=0;ii<8;ii++){
      input[robotNr][ii] += sens[ii];
    }
    break;
  case 4:
    input[robotNr][0] += (sens[0] + sens[1]) / 2;
    input[robotNr][1] += (sens[1] + sens[2]) / 2;
    input[robotNr][2] += (sens[3] + sens[4]) / 2;
    input[robotNr][3] += (sens[4] + sens[5]) / 2;
    input[robotNr][6] += sens[6];
    input[robotNr][7] += sens[7];
    break;
  case 2:
    input[robotNr][0] += (sens[0] + sens[1] + sens[2]) / 3;
    input[robotNr][1] += (sens[3] + sens[4] + sens[5]) / 3;
    input[robotNr][6] += sens[6];
    input[robotNr][7] += sens[7];
    break;
  default:
    printf("Sane check in sim.c removed and Environ::addToInputArray freaked out\n");
    printf("OR\nSomething you got yourself a nice memory leak \n");
    printf("- nrOfFrontInfrared must be 2, 4 or 6 -\n");
    printf("BAILING OUT !!! \n");
    exit(-1);
  }
}

/**
 * Set input array for a robot. Uses global parameters loaded from parameter file. Creates a identical copy in simoutput[][].
 * @param robotNr Which robot to set input for.
 * @todo Environ::setInput() split up function for readability.
 */

void Environ::setInput(int robotNr){
  int i,ii,l;
  float   groundValue=0.0;
  float p1,p2;
  double direction;
  float *sens;
  float radius;

  p1 = robot[robotNr]->getX();
  p2 = robot[robotNr]->getY();
  direction = robot[robotNr]->direction;
  radius = robot[robotNr]->getRadius();
#ifdef ENVIRON_DEBUG
  cout << "Environ::setInput(" << robotNr <<")" << endl;
  cout << "the Robot is at position " << p1 << "x, " << p2 << "y, heading " << direction << endl;
  cout << " * Sensors checking " << nrOfWalls <<" walls" << endl;
#endif
  /*
   * activate wall sensors
   */
  for(i=0;i<NETWORK_INPUTS;i++)
    input[robotNr][i] = 0.0;

  for(i=0;i<nrOfWalls;i++){
    sens=NULL;
    sens = wall[i]->sensorVal(p1,p2,(float)direction,radius);
    if(sens != NULL){
      addToInputArray(robotNr,sens);
    }
  } 

#ifdef ENVIRON_DEBUG
  cout << " * - Wall checking done" << endl;
  cout << " * Sensors checking " << nrOfObst <<" round obstacles" << endl;
#endif
  /*
   * activate robot -> sensors
   */
  if(nrOfRobots>1){
    for(i=0;i<nrOfRobots;i++){
      sens=NULL;
      if(i!=robotNr)
	sens = robot[i]->sensorVal(p1,p2,(float)direction,radius);
      if(sens != NULL){
	addToInputArray(robotNr,sens);
      }
    } 
  }
  
  /*
   * activate sensors from round obstacles
   */
  for(i=0;i<nrOfObst;i++){
    sens=NULL;
    sens = obst[i]->sensorVal(p1,p2,direction,radius);
    if(sens!=NULL){
      addToInputArray(robotNr,sens);
    }
  }

#ifdef ENVIRON_DEBUG
  cout << " * - Round obstacles checking done" << endl;
  cout << " * Sensors checking " << nrOfSObst <<" small round obstacles" << endl;
#endif
  /*
   * activate sensors from small round obstacles if not taken
   */   
  for(i=0;i<nrOfSObst;i++){
    sens=NULL;
    sens = sobst[i]->sensorVal(p1,p2,direction,radius);
    if(sens!=NULL){
      addToInputArray(robotNr,sens);
    } 
  }

#ifdef ENVIRON_DEBUG
  cout << " * - Small round obstacles checking done" << endl;
  if(sensorNoise >0 )
    cout << " * Sensors adding noise " << endl;
#endif


  /*
   * add noise to infra-red sensors if requested
   * if ob_taken == 1 extra noise is present
   */
  if (sensorNoise > 0){
    for(i=0; i< 8; i++){
      if (bot_stat[robotNr]->useGripperSensor == 1 /* && robot[robotNr]->gripper.ob_taken == 1*/) 
	input[robotNr][i] += g_rans(amountOfNoise + 0.0);
      else
	input[robotNr][i] += g_rans(amountOfNoise);	
    }	
  }

  for(ii=0;ii<8;ii++){
    if (input[robotNr][ii] > 1.0)
      input[robotNr][ii] = 1.0;
    if(input[robotNr][ii] < 0.0 )
      input[robotNr][ii] = 0.0;
  }

  /*
   * if we have a sensor of energy we activate it
   */ 
  if (bot_stat[robotNr]->useEnergy){
    input[robotNr][EnergySensor] = robot[robotNr]->energy / 300.0;
    if (input[robotNr][EnergySensor] > 1.0) 
      input[robotNr][EnergySensor] = 1.0;
    else if(input[robotNr][EnergySensor] < 0.0)
      input[robotNr][EnergySensor] = 0.0;
  }   
  //printInput(0);
  /*
   * if we have a ground sensor we activate it
   */ 
  if (bot_stat[robotNr]->useGround){
    groundValue = (float)inZone(robotNr);
    if(groundValue<0){
      input[robotNr][GroundSensor] = 0.0;
    }
    else{
      input[robotNr][GroundSensor] = 1.0;
    }
  }         

  /*
   * if we have a compass sensor we activate it
   */ 
  if (bot_stat[robotNr]->useCompass){
    input[robotNr][CompassSensor] = direction / 360.0;
  }               
  /*
   * activate bulblight sensors  
   */
  if (bot_stat[robotNr]->useLight){ 
    for(i=LightBulbSensor;i<LightBulbSensor+8;i++)
      input[robotNr][i] = 0.0;

    for(i=0;i<nrOfLights;i++){
      sens=NULL;
      sens = light[i]->sensorVal(p1,p2,direction,radius);
      if(sens!=NULL){
	for(ii=LightBulbSensor,l=0;ii< LightBulbSensor + 8; ii++,l++){
	  input[robotNr][ii] += sens[l];
	  if (input[robotNr][ii] > 1.0)
	    input[robotNr][ii] = 1.0;
	}
      }
    } 
    //    printInput(0);
    /*
     * add noise to light sensors if requested
     */ 
    if (sensorNoise == 1){ 
      for(i=LightBulbSensor; i< LightBulbSensor + 8; i++){
	if (input[robotNr][i] > 0.1)
	  input[robotNr][i] += g_rans(input[robotNr][i]);
	else
	  input[robotNr][i]+= g_rans(0.1); 
      if (input[robotNr][i] < 0.0) input[robotNr][i]=0.0;
      if (input[robotNr][i] > 1.0) input[robotNr][i]=1.0;
      }
    }
  }
  
  /*
   * activate gripper sensor
   * No simulation of resistance
   * 
   * Comment stolen from Nolfi's simulator
   * sensore di oggetto della pinza
   */
  if(bot_stat[robotNr]->useGripperSensor){
    if(robot[robotNr]->isaGripperRobot()){
      GripperRobot *bot;
      bot = (GripperRobot*)robot[robotNr];
      input[robotNr][GripperSensor] = bot->ob_sensed;
    }
  }

  /* N */ 
  /*
   * add input from simulated rodcamera
   */
  if(bot_stat[robotNr]->useRodSensor){
    p1 = robot[robotNr]->getX();         
    p2 = robot[robotNr]->getY();          
    direction = robot[robotNr]->direction;
    radius = robot[robotNr]->getRadius(); 
    
    for(i=0; i<nrOfRobots; i++){                                    
      if(i!=robotNr){                                               
        if(robot[i]->nRod>0){                                       
          for(int cRod=0; cRod<robot[i]->nRod; cRod++){             
	    int ra,ka,oa,angle;
	    double oad,ox,oy;
	    // Get X and Y position of Rod
            float x_pos = robot[i]->getX() + 
	      robot[i]->getRodX(cRod)*g_cos(g_DGRtoRAD(robot[i]->direction));
            float y_pos = robot[i]->getY() + 
	      robot[i]->getRodY(cRod)*g_sin(g_DGRtoRAD(robot[i]->direction)); 
	    // Calculate the distance
	     float dist = g_distPoint(p1,p2,x_pos,y_pos);           

	    /* Fix Johan */
	    
	    ox = (double)(x_pos - p1);
	    oy = (double)(y_pos - p2);
	    oad = atan2(ox,oy);
	    oa =(int)g_RADtoDGR(oad);
	    /* coordinate system is upside down */
	    oa = oa * -1;
	    
	    /* rotate kangle to match oa */
	    /* coordinate system is upside down */
	    ka = g_adjustAngle360((int)direction,270);
	    ka = g_splitAngle180(ka);
	    
	    ra =(int)(oa - ka);
	    ra = g_modulo(ra,360);   

	    angle = ra;
            angle = g_modulo(angle+ROD_ANGLE/2, 360);          // To be able to use the angle range [0..CAMERA_ANGLE]
	    // end fix Johan

            if((dist>=0) && (dist<ROD_MAX_DISTANCE) && (angle>=0) && (angle<=ROD_ANGLE)){   // Check if the object is in range and in front of the robot
              // Ok, the object is in range...
              int p_retina = (angle*ROD_RESOLUTION)/ROD_ANGLE;  // Calc the position on the retina
              if((p_retina>=0) && (p_retina<ROD_RESOLUTION))
                input[robotNr][RodSensor+p_retina] = 1;            // Set the corresponding camera position to 1
            }
          }
        }
      }
    }
    
    // If specified, add noise...
    if (useRodNoise == 1){ 
      for(i=RodSensor; i< RodSensor + ROD_RESOLUTION; i++){
        input[robotNr][i] += g_rans(float(amountOfNoise))-amountOfNoise/2;  // Add noise
        if (input[robotNr][i] < 0.0) input[robotNr][i]=0.0;               // Ensure input is between boundaries
        if (input[robotNr][i] > 1.0) input[robotNr][i]=1.0;               // Ensure input is between boundaries
      }
    }
  }
  /* N */
  for(i=0;i < NETWORK_INPUTS; i++){
    simoutput[robotNr][i]=input[robotNr][i];
  }
}
