/*
  YAKS, a Khepera simulator including a separate GA and ANN 
  (Genetic Algoritm, Artificial Neural Net).
  Copyright (C) 2000  Johan Carlsson (johanc@ida.his.se)
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 */

/* Standard libraries */
#include <stdio.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <direct.h>

/* Simulator related includes*/ 
#include "ga.h"
#include "ann.h"
#include "buildn.h"
#include "geom.h"
#include "environ.h"
#include "param.h"
#include "real.h"
#include <sys/stat.h>

#ifndef WIN32
#include <sys/types.h>
#include <signal.h>
#endif
#include "terminal.h"

/* GUI related includes*/
#ifdef GUI
#include "monitor.h"
#include <pthread.h>

#include <support.h>
#include <glib.h>
#include "world.h"
#ifdef HAVE_IMLIB
#include <gdk_imlib.h>
#endif
pthread_t tid;
#ifdef WIN32
#include <Windows.h>
HANDLE gui_run;
HANDLE gui_step;
HANDLE gui_ready;
#else
#include <semaphore.h>
sem_t gui_run;
sem_t gui_step;
sem_t gui_ready;
#endif

                 
#endif /* GUI */

/* The world and such alike */
Environ *world=NULL;

Real *real;


/*  Network, ga and such  */
Ga **nN;
Ann *orginal;
class Ann **annPtr;

int ind;            /* running Individ    */
int gen;            /* running Generation */
int epo;            /* running Epoch      */
int run;            /* running Step (interaction steps between the robot and the world */

#ifdef GUI
Monitor *mon;
#endif

/* double   P90,P180,P360; */
#ifdef GUI
void sim_setWorldSize(){
  gui_setWorldSize((int)(world->dx*xScale),
		   (int)(world->dy*xScale));
}
#endif
void sim_shutdown(){
  exit(0);
}

void sim_saveweights(){
  FILE *fp;
  char tmpStr[400];
  for(int i=0; i < paramNrOfRobots; i++){     
    sprintf(tmpStr,"%s/%s%d-%d.wts",logPath,weightsFile,gen,i);
    if ((fp=fopen(tmpStr, "w")) == NULL){ 
      printf("I can't open weight file %s for writing\n",tmpStr); 
      exit(1); 
    } 
    printf("Saving weights to file %s\n",tmpStr);
    nN[i]->saveAllIndividsWeights(fp);
    fflush(fp);
    fclose(fp);
  }
}

void sigCtrlC(int dontCare){
  Terminal *term;
  term = new Terminal();
  term->addItem("q","Simulator shutdown",sim_shutdown);
  term->addItem("s","Save current generation",sim_saveweights);
#ifndef WIN32
  if(useSerialLine){
    for(int i=0; i < paramNrOfRobots; i++){
      /* Stop the robots */
      fprintf(stderr,"Telling bot %d to stop\n",i);
      real->runStep(0.5,0.5,i);
    }
  }
#endif
  term->runTerminal(); 
  delete(term);
}



int main(int argc,char **argv){

  double *fitness;
  int nrOfInputs,nrOfSensors;
  float *i1,*i2, *i3, *i4;
  float *startPy,*startPx;
  char worldFileStr[200];
  char tmpStr[200];
  int i;

  FILE *fp;
  FILE *actfile;
  seed=time(NULL);
  srand(seed);
#ifndef WIN32
#ifndef __STRICT_ANSI__
  setlinebuf(stdout);
  setlinebuf(stderr);
#endif
#endif

  if(argc > 1)
    loadParams(argv[1]);
  else{
    fprintf(stderr,"No option file specified\n");
    fprintf(stderr,"use: sim optionFile.opt \n\n");
	fprintf(stderr,"trying with default.opt \n\n");
	//char tdf[12] = "default.opt";
	loadParams("default.opt");
    //return -1;
  }
#ifndef WIN32
  signal(SIGHUP,SIG_IGN);
#endif
  i1 = new float[paramNrOfRobots];
  i2 = new float[paramNrOfRobots];
  i3 = new float[paramNrOfRobots];
  i4 = new float[paramNrOfRobots];

  startPy = new float[paramNrOfRobots];
  startPx = new float[paramNrOfRobots];

  annPtr = (class Ann**) malloc(sizeof(class Ann*)*paramNrOfRobots);
  fitness = (double*) malloc(sizeof(double)*paramNrOfRobots);
  nN = (class Ga **) malloc(sizeof(class Ga*)*paramNrOfRobots);
  /* Prepare the world */
  /* First we create the world */
  world = new Environ();
  /* We tell the world to load the world specified in the option file.
   * All necessary sensor sample files will also be loaded.
   */
  sprintf(worldFileStr,"%s/%s",worldPath,worldFile);
  world->loadWorld(worldFileStr);
  /* Add all robots */
  for(int x=0; x < paramNrOfRobots; x++){
    if(robotStr[x]){
      world->addRobot(robotStr[x]);
    }
  }
  for(i=0; i < paramNrOfRobots; i++){
    orginal = new Ann();
    buildNetwork[i%NETWORKS](orginal,world->bot_stat[i]->nrOfSensors);
    /* We create the first generation                  */
    nN[i] = new Ga(nrOfIndivids,amoebaEpochs);
    /* and specifies how the individs ought to look    */
    /* with our precreated neural network              */
    nN[i]->createIndivids(orginal);           
    /* Randomize all weights                           */
#ifdef BIT_WEIGHTS
    nN[i]->mutateIndivids(50.0);
#else
    nN[i]->mutateIndivids(2.0);
#endif
    /* Discard out orginal net we dont need it anymore */
    delete(orginal);
  }


  /** Create a input mapping ANN <-> simulator
   *  (Much of this ought to be done in environ)
   */
  for(int robotNr=0; robotNr < paramNrOfRobots; robotNr++){
    /* What sensor are we going to use, sum them up and compare
     * the number of inputs with the number of sensors  
     */
    nrOfInputs =  nN[robotNr]->individs[0]->getNrOfInputNodes();

    if(world->bot_stat[robotNr]->nrOfSensors > nrOfInputs){
      fprintf(stderr,"Number of ANN inputs doesn't match number of sensors to use on the robot\n");
      fprintf(stderr,"Net inputs %d < %d simulator sensors used\n",nrOfInputs,world->bot_stat[robotNr]->nrOfSensors);
      exit(-1);
    }

    nrOfSensors = 0;
    switch(world->bot_stat[robotNr]->frontInfrared){
    case 6:
      mapping[robotNr][0]=0;
      mapping[robotNr][1]=1;
      mapping[robotNr][2]=2;
      mapping[robotNr][3]=3;
      mapping[robotNr][4]=4;
      mapping[robotNr][5]=5;
      nrOfSensors = 6;
      break;
    case 4:
      mapping[robotNr][0]=0;
      mapping[robotNr][1]=1;
      mapping[robotNr][2]=2;
      mapping[robotNr][3]=3;
      nrOfSensors = 4;
      break;
    case 2:
      mapping[robotNr][0]=0;
      mapping[robotNr][1]=1;
      nrOfSensors = 2;
      break;
    case 0:
      nrOfSensors = 0;
    default:
      printf("Incorrect number of front infrared sensors specified - %d choosed\n",world->bot_stat[robotNr]->frontInfrared);
      printf("Possible choices are 0, 2, 4 or 6\n");
      exit(-1);
      break;
    }
    
    /*
     * Next is the back infrared sensor
     */
    if(world->bot_stat[robotNr]->backInfrared){
      mapping[robotNr][nrOfSensors] = 6;
      mapping[robotNr][nrOfSensors+1] = 7;
      nrOfSensors += 2;
    }

    /*
     *   And then we have all nifty robot sensors
     *
     */
    if(world->bot_stat[robotNr]->useGripperSensor){
      mapping[robotNr][nrOfSensors++] = GripperSensor;
    }
    if(world->bot_stat[robotNr]->useCompass){
      mapping[robotNr][nrOfSensors++] = CompassSensor;
    }
    if(world->bot_stat[robotNr]->useGround){
      mapping[robotNr][nrOfSensors++] = GroundSensor;
    }
    if(world->bot_stat[robotNr]->useEnergy){
      mapping[robotNr][nrOfSensors++] = EnergySensor;
    }
    if(world->bot_stat[robotNr]->useLight){
      for(int l=0; l < 8; l++){
	mapping[robotNr][nrOfSensors+l]=LightBulbSensor+l;
      }
      nrOfSensors+=8;
    }
    if(world->bot_stat[robotNr]->useRodSensor){
      for(int l=0; l < ROD_RESOLUTION; l++){
	mapping[robotNr][nrOfSensors+l]=RodSensor+l;
      }
      nrOfSensors+=ROD_RESOLUTION;

    }
  }

  /* Our ctrl^c handler that starts a terminal */
#ifndef WIN32
  signal(SIGINT,sigCtrlC);  
#endif
#ifdef GUI

  g_thread_init(NULL);

  gtk_set_locale ();
  gtk_init (&argc, &argv);
#ifdef HAVE_IMLIB
  gdk_imlib_init();
#endif
  add_pixmap_directory ((const gchar*)"./pixmaps");
  add_pixmap_directory ((const gchar*)"./pixmaps");

  sim_setWorldSize();
#ifdef WIN32
  gui_ready = CreateSemaphore(NULL,0,1,NULL);
  gui_run = CreateSemaphore(NULL,0,1,NULL);
  gui_step = CreateSemaphore(NULL,1,1,NULL);
#else
  sem_init(&gui_ready,0,0); 
  sem_init(&gui_run,0,0);
  sem_init(&gui_step,0,1);
#endif

  if(pthread_create(&tid, NULL,(void *(*)(void *))gui_init,NULL)!=0){
    perror("main() Error creating GUI thread\n");
    exit(EXIT_FAILURE);
  }

  printf("Wait for gui to launch\n");
#ifdef WIN32
  WaitForSingleObject(gui_ready,INFINITE);
#else
  sem_wait(&gui_ready);
#endif
  printf("GUI ready -> GO !!! \n");
#endif /* GUI */


  /* Some sane tests */
  if(startGeneration > nrOfGenerations){
    fprintf(stderr,"\n\n********\nstartGeneration > nrOfGenerations \n");
    fprintf(stderr,"give me a break man ! I cant go back in time...\n********\n");
    exit(-1);
  }
  if (startGeneration >= 0){
    struct stat tmpS;
    for(i=0; i < paramNrOfRobots; i++){ 
      
      sprintf(tmpStr,"%s/%s%d-%d.wts",logPath,weightsFile,startGeneration-1,i);
      if(stat(tmpStr,&tmpS)!=0){
	if(verboseLevel>1) printf("%s not found, ",tmpStr);
	sprintf(tmpStr,"%s/%s%d-%d.wts",logPath,weightsFile,startGeneration,i);
	if(verboseLevel>1) printf("trying %s instead\n",tmpStr);
	if(stat(tmpStr,&tmpS)!=0){
	  if(verboseLevel>1) printf("failed - ");
	  sprintf(tmpStr,"%s/%s%d-%d.wts",logPath,weightsFile,startGeneration+1,i);
	  if(verboseLevel>1) printf("trying %s instead\n",tmpStr);
	  if(stat(tmpStr,&tmpS)!=0){
	    if(verboseLevel>1) printf("failed\n");
	  }
	}
      }
      if ((fp=fopen(tmpStr, "r")) == NULL){ 
	printf("I can't open generation weights file\n\n"); 
	exit(1); 
      } 
      printf("Loading weights from file %s\n",tmpStr);
      nN[i]->loadAllIndividsWeights(fp);
      
      fclose(fp);
    }
  }
  else{
    startGeneration=0;
	if (_chdir(logPath)) { //if logPath doesn't exist
		printf( "Unable to locate the directory: %s\n", logPath );
		if (_mkdir(logPath)) {
			printf( "Unable to create the directory: %s\n", logPath );
			exit(10);
		} else
			printf( "Directory: %s created\n", logPath );
	} else {
		//delete old log files
		system("del /Q *.*");
	}
  }
  


  if (waitOnStart) {
#ifdef GUI
	printf("Waiting for user to press run\n");
#ifdef WIN32
	WaitForSingleObject(gui_run,INFINITE);
#else
	sem_wait(&gui_run);
#endif
#endif /* GUI */    
  } else {
	gfxUpdate = 0;
  }

  /* monitor */
  //mon = new Monitor(400, 400, 1);
  //mon->addWatch("bla bla",-5,5);
  //mon->createWindow();
  //mon->setWatch(1,1);
  //mon->drawValues();

  /*############################################################REAL###############*/
  /********************************************************************
   *
   *  The main GA loop for startgeneration -> generations
   *
   *
   */

  if(useSerialLine){ /* Real environment */
    real = new Real();
    if(!real->init()){
      fprintf(stderr,"main() init of serial line failed\n");
      exit(-1);
    }
    for(gen=startGeneration;gen<nrOfGenerations;gen++){
      
      if (verboseLevel>0) printf("\nGeneration : %d\n",gen);
      if (verboseLevel>0) printf(" Evaluating primordial soup \n");
      /* Loop trough all individs */
      for(ind=0; ind < nrOfIndivids;ind++){

        if(selectionMethod==-1) {
          /* We just want to watch, so we don't need to calculate a */
          /* correct fitness, nor iterate through all individuals   */
          ind=nrOfIndivids;
        }

	for(i=0; i < paramNrOfRobots; i++){
	  printf(" i%d bot%d\n",ind,i);
	  /* Connect ANN output to robot */
          if(selectionMethod>-1) {
	    annPtr[i] = nN[i]->individs[ind];
          } else {
            /* We just want to watch, so we don't need to calculate a */
            /* correct fitness, nor iterate through all individuals   */
            annPtr[i] = nN[i]->individs[0];  // We know that this is the best individ
	  }
	}
	
	/* The main individ loop for epoch=0 -> amoebaepochs */
	for (epo=0;epo<amoebaEpochs;epo++){

          if(selectionMethod==-1) {
            /* We just want to watch, so we don't need to calculate a */
            /* correct fitness, nor iterate through all individuals   */
            epo=amoebaEpochs;
          }

	  /**************************************************/
	  /*   we reset activation values to 0 in the ANN   */
	  for(i=0; i < paramNrOfRobots; i++)
	    annPtr[i]->resetNet();
	  
	  /**************************************************/
	  /*   Let the robot interact with the environment  */
	  /*   for nrOfSteps time steps                     */
	  for(run = 0; run < nrOfSteps; run++){
	    /**************************************************/
	    /* Let the world generate sensor values for       */
	    /* robot with index 0 -> create inputs values to  */
	    /* ANN   */
	    for(i=0; i < paramNrOfRobots; i++){
	      real->setInput(i); 	 
	      for(int s=0; s < world->bot_stat[i]->nrOfSensors; s++){
		     annPtr[i]->setInput(s,input[i][mapping[i][s]]);
	      }
	      /**************************************************/
	      /* Activate the ANN for individ n                 */
	      /* WARNING the input array for the active robot   */
	      /* will be undefined after this call.             */
	      /* HOWEVER the simoutput array will be stable     */
	      annPtr[i]->activateNet(); 
		      /**************************************************/
	      /* Run the robot with the output from the ANN     */
	      /*                                                */
	      /* If you are going to use a localistic encoding  */
	      /* of the motor actions this is the place to      */
	      /* modify the code.                               */
	      /*                                                */
	      /* runStep expects two values between 0.0 and 1.0 */
	      /* the first for the left motor and the second for*/
	      /* the right motor                                */
	      i1[i]=annPtr[i]->getOutput(0);
	      i2[i]=annPtr[i]->getOutput(1);
	    
	      real->runStep(i1[i],i2[i],i);  // robot 0
#ifdef GUI
	    if(gui_stepping){
#ifdef WIN32
	      WaitForSingleObject(gui_step,INFINITE);
#else
	      sem_wait(&gui_step);
#endif
	    }
	world->drawBots();	
#endif /* GUI */

		}
	    
	    /**************************************************/
	    /* If you want to insert other things that the net*/
	    /* should modify (send it output values to) this  */
	    /* would be a great place do so                   */
	    
	  } /* End of time loop */
	} /* End of epoch loop */
	real->nextIndivid();
      } /* End of individ loop */
    } /* End of generation loop */
  }
  /*############################################################SIMULATED##########*/
  /********************************************************************
   *
   *  The main GA loop for startgeneration -> generations
   *
   *
   */
  else{  /* Simulated environment*/
    for(gen=startGeneration;gen<nrOfGenerations;selectionMethod<0?gen=startGeneration:gen++){
      
      if (verboseLevel>0) printf("\nGeneration : %d\n",gen);
      if (verboseLevel>0) printf(" Running amoeba lives\n");

      /* Loop trough all individs */
      for(ind=0; ind < nrOfIndivids;ind++){
#ifdef GUI
        if(selectionMethod==-1) {
          /* We just want to watch, so we don't need to calculate a */
          /* correct fitness, nor iterate through all individuals   */
          ind=nrOfIndivids;
        }
#endif

		/********************************************************
		*   Connect ANN to robots which will be evaluated
		*
		*/
		for(i=0; i < paramNrOfRobots; i++){
			if(selectionMethod>-1) {
				annPtr[i] = nN[i]->individs[ind];
			} else {
				/* We just want to watch the best individual */
				annPtr[i] = nN[i]->individs[0];
			}
		}
		/* The main individ loop for epoch=0 -> amoebaepochs */
		for (epo=0;epo<amoebaEpochs;epo++){
			/* We just want to watch the best individual */
			if(selectionMethod==-1) {
				epo=amoebaEpochs;
			}
			/***********************************************/
			/* Reset robots, new startpos and such thingis */
			/* make sure  no objects are taken anymore     */
			/* if GUI redraw world                         */
			world->resetEnvironment();
			  
			for(i=0; i < paramNrOfRobots; i++){
				/************************************************/
				/*   we reset activation values to 0 in the ANN   */
				annPtr[i]->resetNet();
				fitness[i] = 0;	  
			  
				/*************************************************
				*   We want to be able to see where the robot was *
				*   in the beginning. So save current status      */
				world->robot[i]->saveState();
				startPx[i]=world->robot[i]->getX();
				startPy[i]=world->robot[i]->getY();
			}
			  
			/**************************************************/
			/* Let the world generate sensor values for the   */
			/* robots                                         */
			for(i=0; i < paramNrOfRobots; i++)
				world->setInput(i); 	 
			/**************************************************/
			/*   Let the robot interact with the environment  */
			/*   for nrOfSteps time steps                     */

			
			/* If last, first or saveGeneration then save activations */
			/* TODO: Perfezionare per tutti i robot!!! */
			if (saveActivations) {
				if (gen%saveGeneration == 0 || gen+1 == nrOfGenerations) {
					sprintf(tmpStr,"%s/%s%d-%d-%d.act",logPath,weightsFile,gen,ind,epo);
					int retries = 0;
					while ((actfile=fopen(tmpStr, "w")) == NULL) {
						printf("I can't open activations file %s for writing\n",tmpStr); 
						if (retries++ > 4) 
							exit(1); 
					} 
				}
			}

			for(run = 0; run < nrOfSteps; run++){

				/*********************************************/
				/* Copy robot activation to networks          */
				for(i=0; i < paramNrOfRobots; i++){
					for(int s=0; s < world->bot_stat[i]->nrOfSensors; s++){
						annPtr[i]->setInput(s,input[i][mapping[i][s]]);
					}
					/**************************************************/
					/* Activate the ANN for individ i                 */
					annPtr[i]->activateNet();
					//annPtr[i]->
					/**************************************************/
					/* Run the robot with the output from the ANN     */
					/*                                                */
					/* If you are going to use a localistic encoding  */
					/* of the motor actions this is the place to      */
					/* modify the code.                               */
					/*                                                */
					/* runStep expects two values between 0.0 and 1.0 */
					/* the first for the left motor and the second for*/
					/* the right motor                                */
					i1[i]=(float)annPtr[i]->getOutput(0);
					i2[i]=(float)annPtr[i]->getOutput(1);
					world->runStep(i1[i],i2[i],i);
					if (world->bot_stat[i]->useGripper)
						world->runGripper(annPtr[i]->getOutput(2),annPtr[i]->getOutput(3),i);
					//mon->setWatch(1, 1);
					//mon->drawValues();
					/**************************************************/
					/* If you want to insert other things that the net*/
					/* should modify (send it output values to) this  */
					/* would be a great place to do so                */
					// do_other_nice_thingy(annPtr[i]->getOutput(3));
				}
				if (saveActivations) {
					if (gen%saveGeneration == 0 || gen+1 == nrOfGenerations) { 
						annPtr[0]->saveActivations(actfile);
					}
				}
#ifdef GUI
				if(gui_stepping){
#ifdef WIN32
					WaitForSingleObject(gui_step,INFINITE);
#else
					sem_wait(&gui_step);
#endif
				}
#endif /* GUI */

				/**************************************************/
				/* Insert your fitness function here              */
				/**************************************************/
			    
				for(i=0; i < paramNrOfRobots; i++){
					double ldist = 10000;
					double ldistt;
					for (int lig=0; lig < world->getNrOfLights(); lig++) {
						ldistt = g_distPoint(world->light[lig]->x, world->light[lig]->y, world->robot[i]->getX(), world->robot[i]->getY());
						ldist = min(ldist, ldistt);
					}

					if(world->checkCrash(i)){   /* If he messed up we dont want him... */
						if(paramNrOfRobots<2) {
							//fitness[i] -= (nrOfSteps - run) * 100;
							fitness[i] -= 1000;
							run=nrOfSteps;   /* so we "end" his life */  
							

						} else {                     /* or give him a bad fitness */
							world->robot[i]->recallState();
							fitness[i] -= 500;
						}
					} else {
						if(ffitness==1) { 
							/*distance from light*/

							if (ldist < 50) 
								fitness[i] += 20;
							else {								
								//fitness[i] += (world->robot[i]->mot1 + world->robot[i]->mot2 - 1) * 8;
								//if ((world->robot[i]->mot1 > 0.7 && world->robot[i]->mot2 > 0.7)) 
								//	fitness[i] += 1000;
								//fitness[i] -= abs(world->robot[i]->mot1 - world->robot[i]->mot2) * 10;
								//fitness[i] -= ldist;
								//if (world->robot[i]->getX() < 20 || world->robot[i]->getX() > world->dx - 20 ||
								//	world->robot[i]->getY() < 20 || world->robot[i]->getY() > world->dy - 20)
								//	fitness[i] -= 30;
								fitness[i] += (world->robot[i]->mot1 + world->robot[i]->mot2 - 1);
								fitness[i] -= abs(world->robot[i]->mot1 - world->robot[i]->mot2) * 10;
							}
						} else if (ffitness == 2) { /* Moving is good */
							fitness[i] += (world->robot[i]->mot1 + world->robot[i]->mot2 - 1);
							fitness[i] -= abs(world->robot[i]->mot1 - world->robot[i]->mot2) * 10;
						}
						
					}
					world->setInput(i); 	 	      
				}
#ifdef GUI	     
				world->drawBots();
//				gui_print2("Fitness: ");
#endif /* GUI */
				/**************************************************/
				/* We want to be able to see where the robot was  */
				/* the last time step. So save current status     */
				/* (in case of a crash in the next timestep for   */
				/* example                                        */
			    
				for(i=0; i < paramNrOfRobots; i++){
					world->robot[i]->saveState();
				}
			} /* End of time loop */

			if (saveActivations) {
				if (gen%saveGeneration == 0 || gen+1 == nrOfGenerations) {
					fflush(actfile);
					fclose(actfile);
				}
			}
#ifdef GUI
			if(selectionMethod>-1){
#endif
				for(i=0; i < paramNrOfRobots; i++){
					/**************************************************/
					/* Set/(add) fitness for individ(n) and  epoch(e) */
					/* with new fitness(fitness)                      */
					if(ffitness==2) { 
						/* euclidian distance between start and stop*/
						double gdist = g_distPoint(startPx[i],startPy[i], world->robot[i]->getX(),world->robot[i]->getY());
						fitness[i] += gdist * 5000;
						//if (gdist < 30)
						//	fitness[i] -= 10000000;
						//if (gdist < 60)
						//	fitness[i] -= 1000000;
						//if (gdist < 90)
						//	fitness[i] -= 100000;
						//if (gdist < 120)
						//	fitness[i] -= 10000;



						//fitness[i] = fitness[i] / 100000;
						//fitness[i] -= g_distPoint(10, 240, world->robot[i]->getX(), world->robot[i]->getY()) * 10000;

					}
					else if(ffitness==1) {
					/* Try to stay under the light   */
						double ldist = 10000;
						double ldistt;
						for (int lig=0; lig<world->getNrOfLights(); lig++) {
							ldistt = g_distPoint(world->light[lig]->x, world->light[lig]->y, world->robot[i]->getX(), world->robot[i]->getY());
							ldist = min(ldist, ldistt);
						}
						//fitness[i] -= ldist * 50000;
						if (ldist < 60 && !world->checkCrash(i)) 
							fitness[i] += 1000;
    				}
					nN[i]->setFitness(ind,epo,fitness[i]);
				}
#ifdef GUI
			}
#endif
			world->resetZones();
			  
		} /* End of epoch loop */
	  } /* End of individ loop */
	      
	  if(verboseLevel>0) printf(" Making mutations\n");
	  if(verboseLevel>0) printf(" Best results : \n");
#ifdef GUI
      /* If we dont just watch a "good" generation */
	  /* Then consider saving the weights          */
	  if(selectionMethod>-1) {
#endif
		  for(i=0; i < paramNrOfRobots; i++) {     
			  /* If last, first or saveGeneration then save weights */
			  if (gen%saveGeneration == 0 || gen+1 == nrOfGenerations) {
				sprintf(tmpStr,"%s/%s%d-%d.wts",logPath,weightsFile,gen,i);
				if ((fp=fopen(tmpStr, "w")) == NULL){ 
					printf("I can't open weight file %s for writing\n",tmpStr); 
					exit(1); 
				} 
				printf("Saving weights to file %s\n",tmpStr);
				nN[i]->saveAllIndividsWeights(fp);
				fflush(fp);
				fclose(fp);
			  }
	      
			/* Save fitness results and close file */
			  if (gen <nrOfGenerations) {
				sprintf(tmpStr,"%s/%s%d",logPath,fitnessFile,i);
				if ((fp=fopen(tmpStr, "a")) == NULL){ 
					printf("I can't open fitness log file %s for appending",tmpStr); 
					exit(1); 
				} 
				nN[i]->saveBestIndividsFitness(nrOfIndividsToLog,fp,verboseLevel);
				fclose(fp);
				
				sprintf(tmpStr,"%s/all%s%d",logPath,fitnessFile,i);
				if ((fp=fopen(tmpStr, "a")) == NULL){ 
					printf("I can't open fitness log file %s for appending",tmpStr); 
					exit(1); 
				} 
				nN[i]->saveAllIndividsFitness(fp);
				fclose(fp);
			  }
		  }
#ifdef GUI
		}
#endif
	      
		/*********************************************************/
		/* We create the next generation (live long and prosper) */
		/* and reset all old fitness.                            */
		/* If you want to use a new selection method change and  */ 
		/* add a case with your selection method (dont forget to */
		/* document it in the ga class).                         */ 
		for(i=0; i < paramNrOfRobots; i++){
			switch(selectionMethod){
#ifdef GUI
				case -1:
					sprintf(tmpStr,"%s/%s%d-%d.wts",logPath,weightsFile,gen+saveGeneration-1,i);
					if ((fp=fopen(tmpStr, "r")) == NULL){ 
						printf("I can't open generation file %s\n\n",tmpStr); 
						exit(1); 
					} 
					printf("Loading weights from file %s\n",tmpStr);
					nN[i]->loadAllIndividsWeights(fp);
			          
					fclose(fp);
					break;
#endif
				case 0:
					nN[i]->createNewGenerationWithWeightMutation(nrOfFathers,nrOfKids,bitMutateProbability);
					break;
				case 1:
					nN[i]->tournamentSelectionWithMutation(bitMutateProbability);
					break;
				case 2:
					nN[i]->tournamentSelectionWithMutationElite(nrOfIndivids-(nrOfFathers*nrOfKids), 1, bitMutateProbability);
					break;
				default:
					nN[i]->createNewGenerationWithWeightMutation(nrOfFathers,nrOfKids,bitMutateProbability);
					break;
			}
			nN[i]->resetFitness();
		}
#ifdef GUI
		if(selectionMethod==-1){
			gen += saveGeneration-1;  // -1 since we always add one to gen every round in the main Generations loop...
		}
#endif
	}
  }
  
#ifdef GUI
  /* Wait for the GUI thread */
  if (waitOnStart) pthread_join(tid,NULL);
#endif
  return(0);
}











