#ifndef WORLD_H
#define WORLD_H
#ifdef WIN32
#pragma warning(disable: 4091)
#endif
#include <stdio.h>
#include <stdlib.h>
#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include "callbacks.h"

#define GUI_COLORS 6
#define GUI_WHITE 0
#define GUI_BLACK 1
#define GUI_BLUE 2
#define GUI_RED 3
#define GUI_GREEN 4
#define GUI_YELLOW 5

typedef struct yWindow{
  GtkWidget *window;
  GdkColormap *cm;
  GdkColor **palette;
  GdkGC *pen;
  GtkDrawingArea *da;
  gint w;
  gint h;
  GdkPixmap *bb;
  yWindow *next;
};



/* Draw */
extern void gui_drawRobot(int x, int y, double direction, int radius,int drawOrErase, int color,int usesRodSensor);
extern void gui_drawWall(int x1, int y1, int x2, int y2);
extern void gui_drawSObst(int x, int y, int radius);
extern void gui_drawObst(int x, int y, int radius);
extern void gui_drawZone(int x, int y, int radius);
extern void gui_drawLight(int x, int y, int radius);
extern void gui_redrawWorld();
extern void gui_recreateWorld();
extern void gui_setWorldSize(int x, int y);
extern void gui_clear();
extern void gui_drawBG();
extern void gui_drawGripper(int *a, int *b, int layer, int color,int c);
extern void gui_updateSensorWindow(int robotNr,int useRodSensorF,int useEnergyF);

#ifdef HAVE_IMLIB
/* Callbacks */
extern void gui_snapshot(GtkWidget *w, gpointer userdata);

/* Pixmap handling */
void gui_save_pixmap(GdkPixmap *pixmap, char *filename);
#endif

/**
 *  Creates a new window 
 *  with specified dimensions in pixels
 *  and title as specified by title.
 *  It has a preallocated  
 *  palette with 256 entries (the same as in the simulator).
 *  @param width The width of the window to create.
 *  @param height The height of the window to create.
 *  @param title The title of the window.
 *  @return The newly created window or NULL if creation failed.
 */
extern yWindow *gui_window_new(gint width, gint height,
			      const gchar *title);

/**
 * Changes the palette entry of window specified with
 * color [0..255] with the new values
 * specified by red, green and blue [0..65536].
 * @param window The window to change palette entry for.
 * @param color Index of the color to change.
 * @param red The amount of red [0..65536].
 * @param green The amount of green [0..65536].
 * @param blue The amount of green [0..65536].
 */
extern void gui_window_palette_change(yWindow *window,
				      gint color,
				      guint red,
				      guint green,
				      guint blue);
/**
 * Sets the current drawing color.
 * @param window The window on which to change current color.
 * @param color The index of the color to choose [0..255].
 */
extern void gui_window_set_color(yWindow *window,
				 gint color);
/**
 * Draws a line between point x1,y1 and x2,y2 on a window.
 * @param window The window to draw on.
 * @param x1 X coordinate of the line start.
 * @param y1 Y coordinate of the line start.
 * @param x2 X coordinate of the line end.
 * @param y2 Y coordinate of the line end.
 */
extern void gui_window_line(yWindow *window,
			    gint x1, gint y1,
			    gint x2, gint y2);
/**
 * Draws a rectangle with the corner
 * x1,y1 and x2,y2. Filled [0..1] where
 * 0 is non filled and 1 is filled.
 * @param window The window to draw on.
 * @param x1 The upper left X coordinate of the rectangle.
 * @param y1 The upper left Y coordinate of the rectangle.
 * @param x2 The lower right X coordinate of the rectangle.
 * @param y2 The lower right Y coordinate of the rectangle.
 * @param filled 1 for filled, 0 for outline.
 */
extern void gui_window_rectangle(yWindow *window,
				 gint x1, gint y1,
				 gint x2, gint y2,
				 gint filled);
/**
 * Draws a circle with center x,y and radius
 * radius. Filled [0..1] where 0 is non filled
 * and 1 is filled.
 * @param window The window to draw on.
 * @param x The X coordinate of the center.
 * @param y The Y coordinate of the center.
 * @param radius The radius of the circle.
 * @param filled If the circle should be filled (0 or 1).
 */
extern void gui_window_circle(yWindow *window,
			      gint x, gint y,
			      gint radius,
			      gint filled);
/**
 * Draws a arc with center x,y from startAngle
 * specified in degrees where 0 degrees is at 
 * 3 o'clock and ends the arc at the endAngle
 * specified relative startAngle.
 * The arc is drawn counter clockwise.
 * Filled [0..1] where 0 is non filled and
 * 1 is filled.
 * @param window The window to draw on.
 * @param x The X coordinate of the center.
 * @param y The Y coordinate of the center.
 * @param radius The radius of the arc.
 * @param filled If the arc should be filled (0 or 1).
 * @param startAngle Start angle of the arc.
 * @param endAngle End angle of the arc.
 */
extern void gui_window_arc(yWindow *window,
			   gint x, gint y,
			   gint radius,
			   gint filled,
			   gint startAngle,
			   gint endAngle);
/**
 * Plots a pixel with the current color
 * at point x,y on a window.
 * @param window The window to plot on.
 * @param x The X coordinate.
 * @param y The X coordinate.
 */
extern void gui_window_plot(yWindow *window,
			    gint x, gint y);
/**
 * Clears the window using the color specified
 * with the index color [0..255].
 * @param window Windor to set current color for.
 * @param color Index of the color to select.
*/

extern void gui_window_clear(yWindow *window,
			     gint color);
/**
 *  Flushes all draws to window to screen.
 *  Call when you want the window to be updated.
 *  @param The window to update
 */			 

extern void gui_window_flush(yWindow *window);

/**
 * Draws text to position x,y with the current selected color on a window..
 * @param win The window to draw on.
 * @param x X coordinate of text start.
 * @param y Y coordinate of text start.
 * @param text The text to draw.
 */

extern void gui_window_text(yWindow *win,gint x, gint y,const char *text);


#endif /* WORLD_H */
